from __future__ import annotations

import pytest

from ansible.plugins.shell.powershell import _parse_clixml, _replace_stderr_clixml, ShellModule

CLIXML_WITH_ERROR = b'#< CLIXML\r\n<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04">' \
    b'<S S="Error">My error</S></Objs>'


def test_replace_stderr_clixml_by_itself():
    data = CLIXML_WITH_ERROR
    expected = b"My error"
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_replace_stderr_clixml_with_pre_and_post_lines():
    data = b"pre\r\n" + CLIXML_WITH_ERROR + b"\r\npost"
    expected = b"pre\r\nMy error\r\npost"
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_replace_stderr_clixml_with_remaining_data_on_line():
    data = b"pre\r\n" + CLIXML_WITH_ERROR + b"inline\r\npost"
    expected = b"pre\r\nMy errorinline\r\npost"
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_replace_stderr_clixml_with_non_utf8_data():
    # \x82 in cp437 is é but is an invalid UTF-8 sequence
    data = CLIXML_WITH_ERROR.replace(b"error", b"\x82rror")
    expected = "My érror".encode("utf-8")
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_replace_stderr_clixml_across_liens():
    data = b"#< CLIXML\r\n<Objs Version=\"foo\">\r\n</Objs>"
    expected = data
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_replace_stderr_clixml_with_invalid_clixml_data():
    data = b"#< CLIXML\r\n<Objs Version=\"foo\"><</Objs>"
    expected = data
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_replace_stderr_clixml_with_no_clixml():
    data = b"foo"
    expected = data
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_replace_stderr_clixml_with_header_but_no_data():
    data = b"foo\r\n#< CLIXML\r\n"
    expected = data
    actual = _replace_stderr_clixml(data)

    assert actual == expected


def test_parse_clixml_empty():
    empty = b'#< CLIXML\r\n<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04"></Objs>'
    expected = b''
    actual = _parse_clixml(empty)
    assert actual == expected


def test_parse_clixml_with_progress():
    progress = b'#< CLIXML\r\n<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04">' \
               b'<Obj S="progress" RefId="0"><TN RefId="0"><T>System.Management.Automation.PSCustomObject</T><T>System.Object</T></TN><MS>' \
               b'<I64 N="SourceId">1</I64><PR N="Record"><AV>Preparing modules for first use.</AV><AI>0</AI><Nil />' \
               b'<PI>-1</PI><PC>-1</PC><T>Completed</T><SR>-1</SR><SD> </SD></PR></MS></Obj></Objs>'
    expected = b''
    actual = _parse_clixml(progress)
    assert actual == expected


def test_parse_clixml_single_stream():
    single_stream = b'#< CLIXML\r\n<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04">' \
                    b'<S S="Error">fake : The term \'fake\' is not recognized as the name of a cmdlet. Check _x000D__x000A_</S>' \
                    b'<S S="Error">the spelling of the name, or if a path was included._x000D__x000A_</S>' \
                    b'<S S="Error">At line:1 char:1_x000D__x000A_</S>' \
                    b'<S S="Error">+ fake cmdlet_x000D__x000A_</S><S S="Error">+ ~~~~_x000D__x000A_</S>' \
                    b'<S S="Error">    + CategoryInfo          : ObjectNotFound: (fake:String) [], CommandNotFoundException_x000D__x000A_</S>' \
                    b'<S S="Error">    + FullyQualifiedErrorId : CommandNotFoundException_x000D__x000A_</S>' \
                    b'<S S="Error"> _x000D__x000A_</S>' \
                    b'</Objs>'
    expected = b"fake : The term 'fake' is not recognized as the name of a cmdlet. Check \r\n" \
               b"the spelling of the name, or if a path was included.\r\n" \
               b"At line:1 char:1\r\n" \
               b"+ fake cmdlet\r\n" \
               b"+ ~~~~\r\n" \
               b"    + CategoryInfo          : ObjectNotFound: (fake:String) [], CommandNotFoundException\r\n" \
               b"    + FullyQualifiedErrorId : CommandNotFoundException\r\n" \
               b" \r\n"
    actual = _parse_clixml(single_stream)
    assert actual == expected


def test_parse_clixml_multiple_streams():
    multiple_stream = b'#< CLIXML\r\n<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04">' \
                      b'<S S="Error">fake : The term \'fake\' is not recognized as the name of a cmdlet. Check _x000D__x000A_</S>' \
                      b'<S S="Error">the spelling of the name, or if a path was included._x000D__x000A_</S>' \
                      b'<S S="Error">At line:1 char:1_x000D__x000A_</S>' \
                      b'<S S="Error">+ fake cmdlet_x000D__x000A_</S><S S="Error">+ ~~~~_x000D__x000A_</S>' \
                      b'<S S="Error">    + CategoryInfo          : ObjectNotFound: (fake:String) [], CommandNotFoundException_x000D__x000A_</S>' \
                      b'<S S="Error">    + FullyQualifiedErrorId : CommandNotFoundException_x000D__x000A_</S><S S="Error"> _x000D__x000A_</S>' \
                      b'<S S="Info">hi info</S>' \
                      b'<S S="Info">other</S>' \
                      b'</Objs>'
    expected = b"hi infoother"
    actual = _parse_clixml(multiple_stream, stream="Info")
    assert actual == expected


def test_parse_clixml_multiple_elements():
    multiple_elements = b'#< CLIXML\r\n#< CLIXML\r\n<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04">' \
                        b'<Obj S="progress" RefId="0"><TN RefId="0"><T>System.Management.Automation.PSCustomObject</T><T>System.Object</T></TN><MS>' \
                        b'<I64 N="SourceId">1</I64><PR N="Record"><AV>Preparing modules for first use.</AV><AI>0</AI><Nil />' \
                        b'<PI>-1</PI><PC>-1</PC><T>Completed</T><SR>-1</SR><SD> </SD></PR></MS></Obj>' \
                        b'<S S="Error">Error 1</S></Objs>' \
                        b'<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04"><Obj S="progress" RefId="0">' \
                        b'<TN RefId="0"><T>System.Management.Automation.PSCustomObject</T><T>System.Object</T></TN><MS>' \
                        b'<I64 N="SourceId">1</I64><PR N="Record"><AV>Preparing modules for first use.</AV><AI>0</AI><Nil />' \
                        b'<PI>-1</PI><PC>-1</PC><T>Completed</T><SR>-1</SR><SD> </SD></PR></MS></Obj>' \
                        b'<Obj S="progress" RefId="1"><TNRef RefId="0" /><MS><I64 N="SourceId">2</I64>' \
                        b'<PR N="Record"><AV>Preparing modules for first use.</AV><AI>0</AI><Nil />' \
                        b'<PI>-1</PI><PC>-1</PC><T>Completed</T><SR>-1</SR><SD> </SD></PR></MS></Obj>' \
                        b'<S S="Error">Error 2</S></Objs>'
    expected = b"Error 1\r\nError 2"
    actual = _parse_clixml(multiple_elements)
    assert actual == expected


@pytest.mark.parametrize('clixml, expected', [
    ('', ''),
    ('just newline _x000A_', 'just newline \n'),
    ('surrogate pair _xD83C__xDFB5_', 'surrogate pair 🎵'),
    ('null char _x0000_', 'null char \0'),
    ('normal char _x0061_', 'normal char a'),
    ('escaped literal _x005F_x005F_', 'escaped literal _x005F_'),
    ('underscope before escape _x005F__x000A_', 'underscope before escape _\n'),
    ('surrogate high _xD83C_', 'surrogate high \uD83C'),
    ('surrogate low _xDFB5_', 'surrogate low \uDFB5'),
    ('lower case hex _x005f_', 'lower case hex _'),
    ('invalid hex _x005G_', 'invalid hex _x005G_'),
    # Tests regex actually matches UTF-16-BE hex chars (b"\x00" then hex char).
    ("_x\u6100\u6200\u6300\u6400_", "_x\u6100\u6200\u6300\u6400_"),
])
def test_parse_clixml_with_comlex_escaped_chars(clixml, expected):
    clixml_data = (
        '<# CLIXML\r\n'
        '<Objs Version="1.1.0.1" xmlns="http://schemas.microsoft.com/powershell/2004/04">'
        f'<S S="Error">{clixml}</S>'
        '</Objs>'
    ).encode()
    b_expected = expected.encode(errors="surrogatepass")

    actual = _parse_clixml(clixml_data)
    assert actual == b_expected


def test_join_path_unc():
    pwsh = ShellModule()
    unc_path_parts = ['\\\\host\\share\\dir1\\\\dir2\\', '\\dir3/dir4', 'dir5', 'dir6\\']
    expected = '\\\\host\\share\\dir1\\dir2\\dir3\\dir4\\dir5\\dir6'
    actual = pwsh.join_path(*unc_path_parts)
    assert actual == expected



# -*- coding: utf-8 -*-
# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


import json
import os
import functools
import pytest
import tempfile

from io import StringIO
from ansible import context
from ansible.cli.galaxy import GalaxyCLI
from ansible.galaxy import api, role, Galaxy
from ansible.module_utils.common.text.converters import to_text
from ansible.utils import context_objects as co


def call_galaxy_cli(args):
    orig = co.GlobalCLIArgs._Singleton__instance
    co.GlobalCLIArgs._Singleton__instance = None
    try:
        return GalaxyCLI(args=['ansible-galaxy', 'role'] + args).run()
    finally:
        co.GlobalCLIArgs._Singleton__instance = orig


@pytest.fixture(autouse=True)
def reset_cli_args():
    co.GlobalCLIArgs._Singleton__instance = None
    yield
    co.GlobalCLIArgs._Singleton__instance = None


@pytest.fixture(autouse=True)
def galaxy_server():
    context.CLIARGS._store = {'ignore_certs': False}
    galaxy_api = api.GalaxyAPI(None, 'test_server', 'https://galaxy.ansible.com')
    return galaxy_api


@pytest.fixture(autouse=True)
def init_role_dir(tmp_path_factory):
    test_dir = to_text(tmp_path_factory.mktemp('test-ÅÑŚÌβŁÈ Roles Input'))
    namespace = 'ansible_namespace'
    role = 'role'
    skeleton_path = os.path.join(os.path.dirname(os.path.split(__file__)[0]), 'cli', 'test_data', 'role_skeleton')
    call_galaxy_cli(['init', '%s.%s' % (namespace, role), '-c', '--init-path', test_dir, '--role-skeleton', skeleton_path])


def mock_NamedTemporaryFile(mocker, **args):
    mock_ntf = mocker.MagicMock()
    mock_ntf.write = mocker.MagicMock()
    mock_ntf.close = mocker.MagicMock()
    mock_ntf.name = None
    return mock_ntf


@pytest.fixture
def init_mock_temp_file(mocker, monkeypatch):
    monkeypatch.setattr(tempfile, 'NamedTemporaryFile', functools.partial(mock_NamedTemporaryFile, mocker))


@pytest.fixture(autouse=True)
def mock_role_download_api(mocker, monkeypatch):
    mock_role_api = mocker.MagicMock()
    mock_role_api.side_effect = [
        StringIO(u''),
    ]
    monkeypatch.setattr(role, 'open_url', mock_role_api)
    return mock_role_api


def test_role_download_github(init_mock_temp_file, mocker, galaxy_server, mock_role_download_api, monkeypatch):
    mock_api = mocker.MagicMock()
    mock_api.side_effect = [
        StringIO(u'{"available_versions":{"v1":"v1/"}}'),
        StringIO(u'{"results":[{"id":"123","github_user":"test_owner","github_repo": "test_role"}]}'),
        StringIO(u'{"results":[{"name": "0.0.1"},{"name": "0.0.2"}]}'),
    ]
    monkeypatch.setattr(api, 'open_url', mock_api)

    role.GalaxyRole(Galaxy(), galaxy_server, 'test_owner.test_role', version="0.0.1").install()

    assert mock_role_download_api.call_count == 1
    assert mock_role_download_api.mock_calls[0][1][0] == 'https://github.com/test_owner/test_role/archive/0.0.1.tar.gz'


def test_role_download_github_default_version(init_mock_temp_file, mocker, galaxy_server, mock_role_download_api, monkeypatch):
    mock_api = mocker.MagicMock()
    mock_api.side_effect = [
        StringIO(u'{"available_versions":{"v1":"v1/"}}'),
        StringIO(u'{"results":[{"id":"123","github_user":"test_owner","github_repo": "test_role"}]}'),
        StringIO(u'{"results":[{"name": "0.0.1"},{"name": "0.0.2"}]}'),
    ]
    monkeypatch.setattr(api, 'open_url', mock_api)

    role.GalaxyRole(Galaxy(), galaxy_server, 'test_owner.test_role').install()

    assert mock_role_download_api.call_count == 1
    assert mock_role_download_api.mock_calls[0][1][0] == 'https://github.com/test_owner/test_role/archive/0.0.2.tar.gz'


def test_role_download_github_no_download_url_for_version(init_mock_temp_file, mocker, galaxy_server, mock_role_download_api, monkeypatch):
    mock_api = mocker.MagicMock()
    mock_api.side_effect = [
        StringIO(u'{"available_versions":{"v1":"v1/"}}'),
        StringIO(u'{"results":[{"id":"123","github_user":"test_owner","github_repo": "test_role"}]}'),
        StringIO(u'{"results":[{"name": "0.0.1"},{"name": "0.0.2","download_url":"http://localhost:8080/test_owner/test_role/0.0.2.tar.gz"}]}'),
    ]
    monkeypatch.setattr(api, 'open_url', mock_api)

    role.GalaxyRole(Galaxy(), galaxy_server, 'test_owner.test_role', version="0.0.1").install()

    assert mock_role_download_api.call_count == 1
    assert mock_role_download_api.mock_calls[0][1][0] == 'https://github.com/test_owner/test_role/archive/0.0.1.tar.gz'


@pytest.mark.parametrize(
    'state,rc',
    [('SUCCESS', 0), ('FAILED', 1),]
)
def test_role_import(state, rc, mocker, galaxy_server, monkeypatch):
    responses = [
        {"available_versions": {"v1": "v1/"}},
        {"results": [{'id': 12345, 'github_user': 'user', 'github_repo': 'role', 'github_reference': None, 'summary_fields': {'role': {'name': 'role'}}}]},
        {"results": [{'state': 'WAITING', 'id': 12345, 'summary_fields': {'task_messages': []}}]},
        {"results": [{'state': state, 'id': 12345, 'summary_fields': {'task_messages': []}}]},
    ]
    mock_api = mocker.MagicMock(side_effect=[StringIO(json.dumps(rsp)) for rsp in responses])
    monkeypatch.setattr(api, 'open_url', mock_api)
    monkeypatch.setattr(GalaxyCLI, '_task_check_delay_sec', 0.1)
    assert call_galaxy_cli(['import', 'user', 'role']) == rc


def test_role_download_url(init_mock_temp_file, mocker, galaxy_server, mock_role_download_api, monkeypatch):
    mock_api = mocker.MagicMock()
    mock_api.side_effect = [
        StringIO(u'{"available_versions":{"v1":"v1/"}}'),
        StringIO(u'{"results":[{"id":"123","github_user":"test_owner","github_repo": "test_role"}]}'),
        StringIO(u'{"results":[{"name": "0.0.1","download_url":"http://localhost:8080/test_owner/test_role/0.0.1.tar.gz"},'
                 u'{"name": "0.0.2","download_url":"http://localhost:8080/test_owner/test_role/0.0.2.tar.gz"}]}'),
    ]
    monkeypatch.setattr(api, 'open_url', mock_api)

    role.GalaxyRole(Galaxy(), galaxy_server, 'test_owner.test_role', version="0.0.1").install()

    assert mock_role_download_api.call_count == 1
    assert mock_role_download_api.mock_calls[0][1][0] == 'http://localhost:8080/test_owner/test_role/0.0.1.tar.gz'


def test_role_download_url_default_version(init_mock_temp_file, mocker, galaxy_server, mock_role_download_api, monkeypatch):
    mock_api = mocker.MagicMock()
    mock_api.side_effect = [
        StringIO(u'{"available_versions":{"v1":"v1/"}}'),
        StringIO(u'{"results":[{"id":"123","github_user":"test_owner","github_repo": "test_role"}]}'),
        StringIO(u'{"results":[{"name": "0.0.1","download_url":"http://localhost:8080/test_owner/test_role/0.0.1.tar.gz"},'
                 u'{"name": "0.0.2","download_url":"http://localhost:8080/test_owner/test_role/0.0.2.tar.gz"}]}'),
    ]
    monkeypatch.setattr(api, 'open_url', mock_api)

    role.GalaxyRole(Galaxy(), galaxy_server, 'test_owner.test_role').install()

    assert mock_role_download_api.call_count == 1
    assert mock_role_download_api.mock_calls[0][1][0] == 'http://localhost:8080/test_owner/test_role/0.0.2.tar.gz'



"""Patches for builtin socket module."""

from __future__ import annotations

import contextlib
import socket
import typing as t

from . import CallablePatch


class _CustomInt(int):
    """Wrapper around `int` to test if subclasses are accepted."""


class GetAddrInfoPatch(CallablePatch):
    """Patch `socket.getaddrinfo` so that its `port` arg works with `int` subclasses."""

    target_container: t.ClassVar = socket
    target_attribute = 'getaddrinfo'

    @classmethod
    def is_patch_needed(cls) -> bool:
        with contextlib.suppress(OSError):
            socket.getaddrinfo('127.0.0.1', _CustomInt(22))
            return False

        return True

    def __call__(self, host, port, *args, **kwargs) -> t.Any:
        if type(port) is not int and isinstance(port, int):  # pylint: disable=unidiomatic-typecheck
            port = int(port)

        return type(self).unpatched_implementation(host, port, *args, **kwargs)



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import pytest

from ansible.module_utils.facts.virtual.sunos import SunOSVirtual


class MockVirtualSysctl(SunOSVirtual):
    def __init__(self, module):
        self.module = module


def mock_get_bin_path(filename):
    cmd_bins = {
        "zonename": "/usr/bin/zonename",
        "virtinfo": "/usr/sbin/virtinfo",
    }
    return cmd_bins.get(filename, None)


def test_get_virtual_facts_global(mocker):
    module = mocker.Mock()
    module.get_bin_path.side_effect = mock_get_bin_path
    module.run_command.return_value = (0, "global", "")
    mixin = MockVirtualSysctl(module=module)
    guest_facts = mixin.get_virtual_facts()
    expected = {
        "virtualization_tech_guest": set(),
        "virtualization_tech_host": set(["zone"]),
    }

    assert guest_facts == expected


@pytest.mark.parametrize(
    ("guest_tech", "expected_guest"),
    [
        pytest.param(
            "VMware",
            "vmware",
            id="VMware",
        ),
        pytest.param(
            "VirtualBox",
            "virtualbox",
            id="VirtualBox",
        ),
    ],
)
def test_get_virtual_facts_guest(mocker, guest_tech, expected_guest):
    module = mocker.Mock()
    module.get_bin_path.side_effect = [
        "/usr/bin/zonename",
        "/usr/sbin/modinfo",
        "/usr/sbin/virtinfo",
    ]
    module.run_command.side_effect = [
        (0, "local", ""),
        (0, guest_tech, ""),
        (0, "", ""),
    ]
    mixin = MockVirtualSysctl(module=module)
    guest_facts = mixin.get_virtual_facts()
    expected = {
        "virtualization_tech_guest": set([expected_guest, "zone"]),
        "virtualization_tech_host": set(),
        "virtualization_type": expected_guest,
        "virtualization_role": "guest",
        "container": "zone",
    }

    assert guest_facts == expected


@pytest.mark.parametrize(
    ("guest_tech", "expected_guest"),
    [
        pytest.param(
            "VMware",
            "vmware",
            id="VMware",
        ),
        pytest.param(
            "VirtualBox",
            "virtualbox",
            id="VirtualBox",
        ),
    ],
)
def test_get_virtual_facts_ldoms(mocker, guest_tech, expected_guest):
    module = mocker.Mock()
    module.get_bin_path.side_effect = [
        "/usr/bin/zonename",
        "/usr/sbin/modinfo",
        "/usr/sbin/virtinfo",
    ]
    module.run_command.side_effect = [
        (0, "local", ""),
        (0, guest_tech, ""),
        (0, "DOMAINROLE|impl=LDoms", ""),
    ]
    mixin = MockVirtualSysctl(module=module)
    guest_facts = mixin.get_virtual_facts()
    expected = {
        "virtualization_tech_guest": set(["ldom", expected_guest, "zone"]),
        "virtualization_tech_host": set(),
        "virtualization_type": "ldom",
        "virtualization_role": "guest",
        "container": "zone",
    }

    assert guest_facts == expected


@pytest.mark.parametrize(
    ("guest_tech", "expected_guest"),
    [
        pytest.param(
            "VMware",
            "vmware",
            id="VMware",
        ),
        pytest.param(
            "VirtualBox",
            "virtualbox",
            id="VirtualBox",
        ),
        pytest.param(
            "Parallels",
            "parallels",
            id="Parallels",
        ),
        pytest.param(
            "HVM domU",
            "xen",
            id="Xen",
        ),
        pytest.param(
            "KVM",
            "kvm",
            id="KVM",
        ),
    ],
)
def test_get_virtual_facts_smbios(mocker, guest_tech, expected_guest):
    module = mocker.Mock()
    module.get_bin_path.side_effect = [
        "/usr/bin/zonename",
        None,
        None,
        "/usr/sbin/smbios",
    ]
    module.run_command.side_effect = [
        (0, "local", ""),
        (0, guest_tech, ""),
    ]
    mixin = MockVirtualSysctl(module=module)
    guest_facts = mixin.get_virtual_facts()
    expected = {
        "virtualization_tech_guest": set([expected_guest, "zone"]),
        "virtualization_tech_host": set(),
        "virtualization_type": expected_guest,
        "virtualization_role": "guest",
        "container": "zone",
    }

    assert guest_facts == expected


def test_get_virtual_facts_openvz(mocker):
    mocker.patch("os.path.exists", return_value=True)
    module = mocker.Mock()
    module.get_bin_path.side_effect = [
        None,  # zonename
        "/usr/sbin/virtinfo",
    ]
    module.run_command.return_value = (0, "", "")
    mixin = MockVirtualSysctl(module=module)
    guest_facts = mixin.get_virtual_facts()
    expected = {
        "virtualization_role": "guest",
        "virtualization_tech_guest": set(["virtuozzo"]),
        "virtualization_tech_host": set(),
        "virtualization_type": "virtuozzo",
    }

    assert guest_facts == expected



"""
Lossy best-effort serialization for Ansible variables.
Default profile for the `to_json` filter.
Deserialization behavior is identical to JSONDecoder, except known Ansible custom serialization markers will raise an error.
"""

from __future__ import annotations as _annotations

import datetime as _datetime
import functools as _functools

from ... import _datatag
from .. import _profiles


class _Profile(_profiles._JSONSerializationProfile["Encoder", "Decoder"]):
    @classmethod
    def post_init(cls) -> None:
        cls.serialize_map = {
            # DTFIX5: support serialization of every type that is supported in the Ansible variable type system
            set: cls.serialize_as_list,
            tuple: cls.serialize_as_list,
            _datetime.date: cls.serialize_as_isoformat,
            _datetime.time: cls.serialize_as_isoformat,
            _datetime.datetime: cls.serialize_as_isoformat,
            # bytes intentionally omitted as they are not a supported variable type, they were not originally supported by the old AnsibleJSONEncoder
            _datatag._AnsibleTaggedDate: cls.discard_tags,
            _datatag._AnsibleTaggedTime: cls.discard_tags,
            _datatag._AnsibleTaggedDateTime: cls.discard_tags,
            _datatag._AnsibleTaggedStr: cls.discard_tags,
            _datatag._AnsibleTaggedInt: cls.discard_tags,
            _datatag._AnsibleTaggedFloat: cls.discard_tags,
            _datatag._AnsibleTaggedSet: cls.discard_tags,
            _datatag._AnsibleTaggedList: cls.discard_tags,
            _datatag._AnsibleTaggedTuple: cls.discard_tags,
            _datatag._AnsibleTaggedDict: cls.discard_tags,
        }

        cls.deserialize_map = {
            '__ansible_unsafe': _functools.partial(cls.unsupported_target_type_error, '__ansible_unsafe'),
            '__ansible_vault': _functools.partial(cls.unsupported_target_type_error, '__ansible_vault'),
        }

        cls.handle_key = cls._handle_key_str_fallback  # type: ignore[method-assign]  # legacy stdlib-compatible key behavior


class Encoder(_profiles.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_profiles.AnsibleProfileJSONDecoder):
    _profile = _Profile



from __future__ import annotations

import collections.abc as c
import itertools
import typing as t

from ansible.errors import AnsibleValueOmittedError, AnsibleError

from ._engine import TemplateEngine


class ChainTemplar:
    """A basic variable layering mechanism that supports templating and obliteration of `omit` values."""

    def __init__(self, *sources: c.Mapping, templar: TemplateEngine) -> None:
        self.sources = sources
        self.templar = templar

    def template(self, key: t.Any, value: t.Any) -> t.Any:
        """
        Render the given value using the templar.
        Intended to be overridden by subclasses.
        """
        return self.templar.template(value)

    def get(self, key: t.Any) -> t.Any:
        """Get the value for the given key, templating the result before returning it."""
        for source in self.sources:
            if key not in source:
                continue

            value = source[key]

            try:
                return self.template(key, value)
            except AnsibleValueOmittedError:
                break  # omit == obliterate - matches historical behavior where dict layers were squashed before templating was applied
            except Exception as ex:
                raise AnsibleError(f'Error while resolving value for {key!r}.', obj=value) from ex

        raise KeyError(key)

    def keys(self) -> t.Iterable[t.Any]:
        """
        Returns a sorted iterable of all keys present in all source layers, without templating associated values.
        Values that resolve to `omit` are thus included.
        """
        return sorted(set(itertools.chain.from_iterable(self.sources)))

    def items(self) -> t.Iterable[t.Tuple[t.Any, t.Any]]:
        """
        Returns a sorted iterable of (key, templated value) tuples.
        Any tuple where the templated value resolves to `omit` will not be included in the result.
        """
        for key in self.keys():
            try:
                yield key, self.get(key)
            except KeyError:
                pass

    def as_dict(self) -> dict[t.Any, t.Any]:
        """Returns a dict representing all layers, squashed and templated, with `omit` values dropped."""
        return dict(self.items())



"""Identify aggregated coverage in one file missing from another."""

from __future__ import annotations

import os
import typing as t

from .....encoding import (
    to_bytes,
)

from .....executor import (
    Delegate,
)

from .....provisioning import (
    prepare_profiles,
)

from . import (
    CoverageAnalyzeTargetsConfig,
    get_target_index,
    make_report,
    read_report,
    write_report,
)

from . import (
    TargetIndexes,
    IndexedPoints,
)


class CoverageAnalyzeTargetsMissingConfig(CoverageAnalyzeTargetsConfig):
    """Configuration for the `coverage analyze targets missing` command."""

    def __init__(self, args: t.Any) -> None:
        super().__init__(args)

        self.from_file: str = args.from_file
        self.to_file: str = args.to_file
        self.output_file: str = args.output_file

        self.only_gaps: bool = args.only_gaps
        self.only_exists: bool = args.only_exists


def command_coverage_analyze_targets_missing(args: CoverageAnalyzeTargetsMissingConfig) -> None:
    """Identify aggregated coverage in one file missing from another."""
    host_state = prepare_profiles(args)  # coverage analyze targets missing

    if args.delegate:
        raise Delegate(host_state=host_state)

    from_targets, from_path_arcs, from_path_lines = read_report(args.from_file)
    to_targets, to_path_arcs, to_path_lines = read_report(args.to_file)
    target_indexes: TargetIndexes = {}

    if args.only_gaps:
        arcs = find_gaps(from_path_arcs, from_targets, to_path_arcs, target_indexes, args.only_exists)
        lines = find_gaps(from_path_lines, from_targets, to_path_lines, target_indexes, args.only_exists)
    else:
        arcs = find_missing(from_path_arcs, from_targets, to_path_arcs, to_targets, target_indexes, args.only_exists)
        lines = find_missing(from_path_lines, from_targets, to_path_lines, to_targets, target_indexes, args.only_exists)

    report = make_report(target_indexes, arcs, lines)
    write_report(args, report, args.output_file)


def find_gaps(
    from_data: IndexedPoints,
    from_index: list[str],
    to_data: IndexedPoints,
    target_indexes: TargetIndexes,
    only_exists: bool,
) -> IndexedPoints:
    """Find gaps in coverage between the from and to data sets."""
    target_data: IndexedPoints = {}

    for from_path, from_points in from_data.items():
        if only_exists and not os.path.isfile(to_bytes(from_path)):
            continue

        to_points = to_data.get(from_path, {})

        gaps = set(from_points.keys()) - set(to_points.keys())

        if gaps:
            gap_points = dict((key, value) for key, value in from_points.items() if key in gaps)
            target_data[from_path] = dict((gap, set(get_target_index(from_index[i], target_indexes) for i in indexes)) for gap, indexes in gap_points.items())

    return target_data


def find_missing(
    from_data: IndexedPoints,
    from_index: list[str],
    to_data: IndexedPoints,
    to_index: list[str],
    target_indexes: TargetIndexes,
    only_exists: bool,
) -> IndexedPoints:
    """Find coverage in from_data not present in to_data (arcs or lines)."""
    target_data: IndexedPoints = {}

    for from_path, from_points in from_data.items():
        if only_exists and not os.path.isfile(to_bytes(from_path)):
            continue

        to_points = to_data.get(from_path, {})

        for from_point, from_target_indexes in from_points.items():
            to_target_indexes = to_points.get(from_point, set())

            remaining_targets = set(from_index[i] for i in from_target_indexes) - set(to_index[i] for i in to_target_indexes)

            if remaining_targets:
                target_index = target_data.setdefault(from_path, {}).setdefault(from_point, set())
                target_index.update(get_target_index(name, target_indexes) for name in remaining_targets)

    return target_data



#!/usr/bin/python
from __future__ import annotations

import json

DOCUMENTATION = r"""
module: testmodule
description: for testing
extends_documentation_fragment:
  - testns.testcoll.frag
  - testns.testcoll.frag.other_documentation
"""


def main():
    print(json.dumps(dict(changed=False, source='user')))


if __name__ == '__main__':
    main()



# (c) 2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2018, Ansible Project
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import copy
import os
import tempfile
import time
import typing as t

from abc import abstractmethod
from collections import abc as c

from ansible import constants as C
from ansible.errors import AnsibleError
from ansible.module_utils.common.file import S_IRWU_RG_RO
from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins import AnsiblePlugin
from ansible.plugins.loader import cache_loader
from ansible.utils.collection_loader import resource_from_fqcr
from ansible.utils.display import Display

display = Display()


class BaseCacheModule(AnsiblePlugin):

    # Backwards compat only.  Just import the global display instead
    _display = display
    _persistent = True
    """Plugins that do not persist data between runs can set False to bypass schema-version key munging and JSON serialization wrapper."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

        self.set_options(var_options=args, direct=kwargs)

    @abstractmethod
    def get(self, key: str) -> dict[str, object]:
        pass

    @abstractmethod
    def set(self, key: str, value: dict[str, object]) -> None:
        pass

    @abstractmethod
    def keys(self) -> t.Sequence[str]:
        pass

    @abstractmethod
    def contains(self, key: object) -> bool:
        pass

    @abstractmethod
    def delete(self, key: str) -> None:
        pass

    @abstractmethod
    def flush(self) -> None:
        pass


class BaseFileCacheModule(BaseCacheModule):
    """
    A caching module backed by file based storage.
    """
    def __init__(self, *args, **kwargs):

        try:
            super(BaseFileCacheModule, self).__init__(*args, **kwargs)
            self._cache_dir = self._get_cache_connection(self.get_option('_uri'))
            self._timeout = float(self.get_option('_timeout'))
        except KeyError:
            self._cache_dir = self._get_cache_connection(C.CACHE_PLUGIN_CONNECTION)
            self._timeout = float(C.CACHE_PLUGIN_TIMEOUT)
        self.plugin_name = resource_from_fqcr(self.__module__)
        self._cache = {}
        self.validate_cache_connection()

    def _get_cache_connection(self, source):
        if source:
            try:
                return os.path.expanduser(os.path.expandvars(source))
            except TypeError:
                pass

    def validate_cache_connection(self):
        if not self._cache_dir:
            raise AnsibleError("error, '%s' cache plugin requires the 'fact_caching_connection' config option "
                               "to be set (to a writeable directory path)" % self.plugin_name)

        if not os.path.exists(self._cache_dir):
            try:
                os.makedirs(self._cache_dir)
            except OSError as ex:
                raise AnsibleError(f"Error in {self.plugin_name!r} cache plugin while trying to create cache dir {self._cache_dir!r}.") from ex
        else:
            for x in (os.R_OK, os.W_OK, os.X_OK):
                if not os.access(self._cache_dir, x):
                    raise AnsibleError("error in '%s' cache, configured path (%s) does not have necessary permissions (rwx), disabling plugin" % (
                        self.plugin_name, self._cache_dir))

    def _get_cache_file_name(self, key: str) -> str:
        prefix = self.get_option('_prefix')
        if prefix:
            cachefile = "%s/%s%s" % (self._cache_dir, prefix, key)
        else:
            cachefile = "%s/%s" % (self._cache_dir, key)
        return cachefile

    def get(self, key):
        """ This checks the in memory cache first as the fact was not expired at 'gather time'
        and it would be problematic if the key did expire after some long running tasks and
        user gets 'undefined' error in the same play """

        if key not in self._cache:

            if self.has_expired(key) or key == "":
                raise KeyError

            cachefile = self._get_cache_file_name(key)
            try:
                value = self._load(cachefile)
                self._cache[key] = value
            except ValueError as e:
                display.warning("error in '%s' cache plugin while trying to read %s : %s. "
                                "Most likely a corrupt file, so erasing and failing." % (self.plugin_name, cachefile, to_bytes(e)))
                self.delete(key)
                raise AnsibleError("The cache file %s was corrupt, or did not otherwise contain valid data. "
                                   "It has been removed, so you can re-run your command now." % cachefile)
            except FileNotFoundError:
                raise KeyError
            except Exception as ex:
                raise AnsibleError(f"Error while accessing the cache file {cachefile!r}.") from ex

        return self._cache.get(key)

    def set(self, key, value):

        self._cache[key] = value

        cachefile = self._get_cache_file_name(key)
        tmpfile_handle, tmpfile_path = tempfile.mkstemp(dir=os.path.dirname(cachefile))
        try:
            try:
                self._dump(value, tmpfile_path)
            except OSError as ex:
                display.error_as_warning(f"Error in {self.plugin_name!r} cache plugin while trying to write to {tmpfile_path!r}.", exception=ex)
            try:
                os.rename(tmpfile_path, cachefile)
                os.chmod(cachefile, mode=S_IRWU_RG_RO)
            except OSError as ex:
                display.error_as_warning(f"Error in {self.plugin_name!r} cache plugin while trying to move {tmpfile_path!r} to {cachefile!r}.", exception=ex)
        finally:
            try:
                os.unlink(tmpfile_path)
            except OSError:
                pass

    def has_expired(self, key):

        if self._timeout == 0:
            return False

        cachefile = self._get_cache_file_name(key)
        try:
            st = os.stat(cachefile)
        except FileNotFoundError:
            return False
        except OSError as ex:
            display.error_as_warning(f"Error in {self.plugin_name!r} cache plugin while trying to stat {cachefile!r}.", exception=ex)

            return False

        if time.time() - st.st_mtime <= self._timeout:
            return False

        if key in self._cache:
            del self._cache[key]
        return True

    def keys(self):
        # When using a prefix we must remove it from the key name before
        # checking the expiry and returning it to the caller. Keys that do not
        # share the same prefix cannot be fetched from the cache.
        prefix = self.get_option('_prefix')
        prefix_length = len(prefix)
        keys = []
        for k in os.listdir(self._cache_dir):
            if k.startswith('.') or not k.startswith(prefix):
                continue

            k = k[prefix_length:]
            if not self.has_expired(k):
                keys.append(k)

        return keys

    def contains(self, key):
        cachefile = self._get_cache_file_name(key)

        if key in self._cache:
            return True

        if self.has_expired(key):
            return False
        try:
            os.stat(cachefile)
            return True
        except FileNotFoundError:
            return False
        except OSError as ex:
            display.error_as_warning(f"Error in {self.plugin_name!r} cache plugin while trying to stat {cachefile!r}.", exception=ex)

    def delete(self, key):
        try:
            del self._cache[key]
        except KeyError:
            pass
        try:
            os.remove(self._get_cache_file_name(key))
        except OSError:
            pass  # TODO: only pass on non existing?

    def flush(self):
        self._cache = {}
        for key in self.keys():
            self.delete(key)

    @abstractmethod
    def _load(self, filepath: str) -> object:
        """
        Read data from a filepath and return it as a value

        :arg filepath: The filepath to read from.
        :returns: The value stored in the filepath

        This method reads from the file on disk and takes care of any parsing
        and transformation of the data before returning it.  The value
        returned should be what Ansible would expect if it were uncached data.

        .. note:: Filehandles have advantages but calling code doesn't know
            whether this file is text or binary, should be decoded, or accessed via
            a library function.  Therefore the API uses a filepath and opens
            the file inside of the method.
        """
        pass

    @abstractmethod
    def _dump(self, value: object, filepath: str) -> None:
        """
        Write data to a filepath

        :arg value: The value to store
        :arg filepath: The filepath to store it at
        """
        pass


class CachePluginAdjudicator(c.MutableMapping):
    """Batch update wrapper around a cache plugin."""

    def __init__(self, plugin_name='memory', **kwargs):
        self._cache = {}
        self._retrieved = {}
        self._plugin = cache_loader.get(plugin_name, **kwargs)

    def update_cache_if_changed(self):
        if self._retrieved != self._cache:
            self.set_cache()

    def set_cache(self):
        for top_level_cache_key in self._cache.keys():
            self._plugin.set(top_level_cache_key, self._cache[top_level_cache_key])

        self._retrieved = copy.deepcopy(self._cache)

    def load_whole_cache(self):
        for key in self._plugin.keys():
            self._cache[key] = self._plugin.get(key)

    def __repr__(self):
        return repr(self._cache)

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    def _do_load_key(self, key):
        load = False

        if key not in self._cache and key not in self._retrieved and self._plugin._persistent and self._plugin.contains(key):
            load = True

        return load

    def __getitem__(self, key):
        if self._do_load_key(key):
            try:
                self._cache[key] = self._plugin.get(key)
            except KeyError:
                pass
            else:
                self._retrieved[key] = self._cache[key]

        return self._cache[key]

    def get(self, key, default=None):
        if self._do_load_key(key):
            try:
                self._cache[key] = self._plugin.get(key)
            except KeyError:
                pass
            else:
                self._retrieved[key] = self._cache[key]

        return self._cache.get(key, default)

    def items(self):
        return self._cache.items()

    def values(self):
        return self._cache.values()

    def keys(self):
        return self._cache.keys()

    def pop(self, key, *args):
        if args:
            return self._cache.pop(key, args[0])

        return self._cache.pop(key)

    def __delitem__(self, key):
        del self._cache[key]

    def __setitem__(self, key, value):
        self._cache[key] = value

    def clear(self):
        self.flush()

    def flush(self):
        self._plugin.flush()
        self._cache = {}

    def update(self, value):
        self._cache.update(value)



"""Command line parsing for the `integration` command."""

from __future__ import annotations

import argparse
import collections.abc as c
import typing as t

from ....commands.integration.posix import (
    command_posix_integration,
)

from ....config import (
    PosixIntegrationConfig,
)

from ....target import (
    walk_posix_integration_targets,
)

from ...environments import (
    CompositeActionCompletionFinder,
    ControllerMode,
    TargetMode,
    add_environments,
)


def do_posix_integration(
    subparsers,
    parent: argparse.ArgumentParser,
    add_integration_common: c.Callable[[argparse.ArgumentParser], None],
    completer: CompositeActionCompletionFinder,
):
    """Command line parsing for the `integration` command."""
    parser: argparse.ArgumentParser = subparsers.add_parser(
        'integration',
        parents=[parent],
        help='posix integration tests',
    )

    parser.set_defaults(
        func=command_posix_integration,
        targets_func=walk_posix_integration_targets,
        config=PosixIntegrationConfig,
    )

    posix_integration = t.cast(argparse.ArgumentParser, parser.add_argument_group(title='integration test arguments'))

    add_integration_common(posix_integration)

    add_environments(parser, completer, ControllerMode.DELEGATED, TargetMode.POSIX_INTEGRATION)  # integration



# (c) 2017, ansible by Red Hat
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

# moved actual classes to __init__ kept here for backward compat with 3rd parties
from ansible.plugins.cache import BaseCacheModule, BaseFileCacheModule  # pylint: disable=unused-import

from ansible.utils.display import Display as _Display

_Display().deprecated(
    msg="The `ansible.plugins.cache.base` Python module is deprecated.",
    help_text="Import from `ansible.plugins.cache` instead.",
    version="2.23",
)



# -*- coding: utf-8 -*-
# (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest
import typing as t

from ansible.module_utils._internal import _traceback, _messages
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common import warnings

from ansible.module_utils.common.warnings import warn
from ansible.module_utils.testing import patch_module_args
from units.mock.module import ModuleEnvMocker

pytestmark = pytest.mark.usefixtures("as_target", "module_env_mocker")


def test_warn():
    warn('Warning message')
    assert warnings.get_warning_messages() == ('Warning message',)
    assert warnings.get_warnings() == [_messages.WarningSummary(event=_messages.Event(msg='Warning message'))]


def test_warn_via_module() -> None:
    with patch_module_args():
        am = AnsibleModule(argument_spec={})

    am.warn('Warning message')

    assert warnings.get_warning_messages() == ('Warning message',)
    assert warnings.get_warnings() == [_messages.WarningSummary(event=_messages.Event(msg='Warning message'))]


def test_multiple_warnings():
    messages = [
        'First warning',
        'Second warning',
        'Third warning',
    ]

    for w in messages:
        warn(w)

    assert warnings.get_warning_messages() == tuple(messages)
    assert warnings.get_warnings() == [_messages.WarningSummary(event=_messages.Event(msg=w)) for w in messages]


def test_dedupe_with_traceback(module_env_mocker: ModuleEnvMocker) -> None:
    module_env_mocker.set_traceback_config([_traceback.TracebackEvent.WARNING])
    msg = "a warning message"

    # WarningSummary dataclass object hash is the dedupe key; presence of differing tracebacks or SourceContexts affects de-dupe

    for _i in range(0, 10):
        warn(msg)  # same location, same traceback- should be collapsed to one message

    assert len(warnings.get_warning_messages()) == 1
    assert len(warnings.get_warnings()) == 1

    for _i in range(0, 10):
        warn(msg)  # with tracebacks on, we should have a different source location than the first loop, but still de-dupe

    assert len(warnings.get_warning_messages()) == 2
    assert len(warnings.get_warnings()) == 2


@pytest.mark.parametrize(
    'test_case',
    (
        1,
        True,
        [1],
        {'k1': 'v1'},
        (1, 2),
        6.62607004,
        b'bytestr',
        None,
    )
)
def test_warn_failure(test_case: t.Any):
    with pytest.raises(TypeError, match=f"must be <class 'str'> instead of {type(test_case)}"):
        warn(test_case)



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

#############################################
from __future__ import annotations

import sys
import typing as t

from ansible import constants as C
from ansible.errors import AnsibleError
from ansible.inventory.group import Group
from ansible.inventory.host import Host
from ansible.utils.display import Display
from ansible.utils.vars import combine_vars
from ansible.utils.path import basedir

from . import helpers  # this is left as a module import to facilitate easier unit test patching


display = Display()


class InventoryData:
    """
    Holds inventory data (host and group objects).
    Using its methods should guarantee expected relationships and data.
    """

    def __init__(self) -> None:

        self.groups: dict[str, Group] = {}
        self.hosts: dict[str, Host] = {}

        # provides 'groups' magic var, host object has group_names
        self._groups_dict_cache: dict[str, list[str]] = {}

        # current localhost, implicit or explicit
        self.localhost: Host | None = None

        self.current_source: str | None = None
        self.processed_sources: list[str] = []

        # Always create the 'all' and 'ungrouped' groups,
        for group in ('all', 'ungrouped'):
            self.add_group(group)

        self.add_child('all', 'ungrouped')

    def _create_implicit_localhost(self, pattern: str) -> Host:

        if self.localhost:
            new_host = self.localhost
        else:
            new_host = Host(pattern)

            new_host.address = "127.0.0.1"
            new_host.implicit = True

            # set localhost defaults
            py_interp = sys.executable
            if not py_interp:
                # sys.executable is not set in some cornercases. see issue #13585
                py_interp = '/usr/bin/python'
                display.warning('Unable to determine python interpreter from sys.executable. Using /usr/bin/python default. '
                                'You can correct this by setting ansible_python_interpreter for localhost')
            new_host.set_variable("ansible_python_interpreter", py_interp)
            new_host.set_variable("ansible_connection", 'local')

            self.localhost = new_host

        return new_host

    def reconcile_inventory(self) -> None:
        """Ensure inventory basic rules, run after updates."""

        display.debug('Reconcile groups and hosts in inventory.')
        self.current_source = None

        group_names = set()
        # set group vars from group_vars/ files and vars plugins
        for g in self.groups:
            group = self.groups[g]
            group_names.add(group.name)

            # ensure all groups inherit from 'all'
            if group.name != 'all' and not group.get_ancestors():
                self.add_child('all', group.name)

        host_names = set()
        # get host vars from host_vars/ files and vars plugins
        for host in self.hosts.values():
            host_names.add(host.name)

            mygroups = host.get_groups()

            if self.groups['ungrouped'] in mygroups:
                # clear ungrouped of any incorrectly stored by parser
                if set(mygroups).difference({self.groups['all'], self.groups['ungrouped']}):
                    self.groups['ungrouped'].remove_host(host)

            elif not host.implicit:
                # add ungrouped hosts to ungrouped, except implicit
                length = len(mygroups)
                if length == 0 or (length == 1 and self.groups['all'] in mygroups):
                    self.add_child('ungrouped', host.name)

            # special case for implicit hosts
            if host.implicit:
                host.vars = combine_vars(self.groups['all'].get_vars(), host.vars)

        # warn if overloading identifier as both group and host
        for conflict in group_names.intersection(host_names):
            display.warning("Found both group and host with same name: %s" % conflict)

        self._groups_dict_cache = {}

    def get_host(self, hostname: str) -> Host | None:
        """Fetch host object using name deal with implicit localhost."""

        hostname = helpers.remove_trust(hostname)

        matching_host = self.hosts.get(hostname, None)

        # if host is not in hosts dict
        if matching_host is None and hostname in C.LOCALHOST:
            # might need to create implicit localhost
            matching_host = self._create_implicit_localhost(hostname)

        return matching_host

    def add_group(self, group: str) -> str:
        """Adds a group to inventory if not there already, returns named actually used."""

        if group:
            if not isinstance(group, str):
                raise AnsibleError("Invalid group name supplied, expected a string but got %s for %s" % (type(group), group))
            if group not in self.groups:
                g = Group(group)
                group = g.name  # the group object may have sanitized the group name; use whatever it has
                if group not in self.groups:
                    self.groups[group] = g
                    self._groups_dict_cache = {}
                    display.debug("Added group %s to inventory" % group)
            else:
                display.debug("group %s already in inventory" % group)
        else:
            raise AnsibleError("Invalid empty/false group name provided: %s" % group)

        return group

    def remove_group(self, group: Group) -> None:

        if group.name in self.groups:
            del self.groups[group.name]
            display.debug("Removed group %s from inventory" % group.name)
            self._groups_dict_cache = {}

        for host in self.hosts:
            h = self.hosts[host]
            h.remove_group(group)

    def add_host(self, host: str, group: str | None = None, port: int | str | None = None) -> str:
        """Adds a host to inventory and possibly a group if not there already."""

        host = helpers.remove_trust(host)

        if host:
            if not isinstance(host, str):
                raise AnsibleError("Invalid host name supplied, expected a string but got %s for %s" % (type(host), host))

            # TODO: add to_safe_host_name
            g = None
            if group:
                if group in self.groups:
                    g = self.groups[group]
                else:
                    raise AnsibleError("Could not find group %s in inventory" % group)

            if host not in self.hosts:
                h = Host(host, port)
                self.hosts[host] = h
                if self.current_source:  # set to 'first source' in which host was encountered
                    self.set_variable(host, 'inventory_file', self.current_source)
                    self.set_variable(host, 'inventory_dir', basedir(self.current_source))
                else:
                    self.set_variable(host, 'inventory_file', None)
                    self.set_variable(host, 'inventory_dir', None)
                display.debug("Added host %s to inventory" % host)

                # set default localhost from inventory to avoid creating an implicit one. Last localhost defined 'wins'.
                if host in C.LOCALHOST:
                    if self.localhost is None:
                        self.localhost = self.hosts[host]
                        display.vvvv("Set default localhost to %s" % h)
                    else:
                        display.warning("A duplicate localhost-like entry was found (%s). First found localhost was %s" % (h, self.localhost.name))
            else:
                h = self.hosts[host]

            if g:
                g.add_host(h)
                self._groups_dict_cache = {}
                display.debug("Added host %s to group %s" % (host, group))
        else:
            raise AnsibleError("Invalid empty host name provided: %s" % host)

        return host

    def remove_host(self, host: Host) -> None:

        if host.name in self.hosts:
            del self.hosts[host.name]

        for group in self.groups:
            g = self.groups[group]
            g.remove_host(host)

    def set_variable(self, entity: str, varname: str, value: t.Any) -> None:
        """Sets a variable for an inventory object."""

        inv_object: Host | Group

        if entity in self.groups:
            inv_object = self.groups[entity]
        elif entity in self.hosts:
            inv_object = self.hosts[entity]
        else:
            raise AnsibleError("Could not identify group or host named %s" % entity)

        inv_object.set_variable(varname, value)
        display.debug('set %s for %s' % (varname, entity))

    def add_child(self, group: str, child: str) -> bool:
        """Add host or group to group."""
        if group in self.groups:
            g = self.groups[group]
            if child in self.groups:
                added = g.add_child_group(self.groups[child])
            elif child in self.hosts:
                added = g.add_host(self.hosts[child])
            else:
                raise AnsibleError("%s is not a known host nor group" % child)
            self._groups_dict_cache = {}
            display.debug('Group %s now contains %s' % (group, child))
        else:
            raise AnsibleError("%s is not a known group" % group)
        return added

    def get_groups_dict(self) -> dict[str, list[str]]:
        """
        We merge a 'magic' var 'groups' with group name keys and hostname list values into every host variable set. Cache for speed.
        """
        if not self._groups_dict_cache:
            for group_name, group in self.groups.items():
                self._groups_dict_cache[group_name] = [h.name for h in group.get_hosts()]

        return self._groups_dict_cache



from __future__ import annotations

from ansible._internal._ansiballz._builder import ExtensionManager
from ansible.module_utils._internal._ansiballz._extensions import _pydevd


def test_debugger_source_mapping() -> None:
    """Synthetic coverage for builder source mapping."""
    debug_options = _pydevd.Options(source_mapping={
        "ide/path.py": "controller/path.py",
        "ide/something.py": "controller/not_match.py",
    })

    manager = ExtensionManager(debug_options)
    manager.source_mapping.update({
        "controller/path.py": "zip/path.py",
        "controller/other.py": "not_match.py",
    })

    extensions = manager.get_extensions()

    assert extensions['_pydevd']['source_mapping'] == {'controller/other.py': 'not_match.py', 'ide/path.py': 'zip/path.py'}



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.module_utils.common.text.converters import to_native
from ansible.module_utils.common.validation import check_mutually_exclusive


@pytest.fixture
def mutually_exclusive_terms():
    return [
        ('string1', 'string2',),
        ('box', 'fox', 'socks'),
    ]


def test_check_mutually_exclusive(mutually_exclusive_terms):
    params = {
        'string1': 'cat',
        'fox': 'hat',
    }
    assert check_mutually_exclusive(mutually_exclusive_terms, params) == []


def test_check_mutually_exclusive_found(mutually_exclusive_terms):
    params = {
        'string1': 'cat',
        'string2': 'hat',
        'fox': 'red',
        'socks': 'blue',
    }
    expected = "parameters are mutually exclusive: string1|string2, box|fox|socks"

    with pytest.raises(TypeError) as e:
        check_mutually_exclusive(mutually_exclusive_terms, params)

    assert to_native(e.value) == expected


def test_check_mutually_exclusive_none():
    terms = None
    params = {
        'string1': 'cat',
        'fox': 'hat',
    }
    assert check_mutually_exclusive(terms, params) == []


def test_check_mutually_exclusive_no_params(mutually_exclusive_terms):
    with pytest.raises(TypeError) as te:
        check_mutually_exclusive(mutually_exclusive_terms, None)
    assert "'NoneType' object is not iterable" in to_native(te.value)



"""Ansible-specific pylint plugin for checking deprecation comments."""

# (c) 2018, Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import shlex
import tokenize

import pylint.checkers
import pylint.lint

import ansible.release

from ansible.module_utils.compat.version import LooseVersion


class AnsibleDeprecatedCommentChecker(pylint.checkers.BaseTokenChecker):
    """Checks for ``# deprecated:`` comments to ensure that the ``version`` has not passed or met the time for removal."""

    name = 'deprecated-comment'
    msgs = {
        'E9601': (
            "Deprecated core version (%r) found: %s",
            "ansible-deprecated-version-comment",
            None,
        ),
        'E9602': (
            "Deprecated comment contains invalid keys %r",
            "ansible-deprecated-version-comment-invalid-key",
            None,
        ),
        'E9603': (
            "Deprecated comment missing version",
            "ansible-deprecated-version-comment-missing-version",
            None,
        ),
        'E9604': (
            "Deprecated python version (%r) found: %s",
            "ansible-deprecated-python-version-comment",
            None,
        ),
        'E9605': (
            "Deprecated comment contains invalid version %r: %s",
            "ansible-deprecated-version-comment-invalid-version",
            None,
        ),
    }

    ANSIBLE_VERSION = LooseVersion('.'.join(ansible.release.__version__.split('.')[:3]))
    """The current ansible-core X.Y.Z version."""

    def process_tokens(self, tokens: list[tokenize.TokenInfo]) -> None:
        for token in tokens:
            if token.type == tokenize.COMMENT:
                self._process_comment(token)

    def _deprecated_string_to_dict(self, token: tokenize.TokenInfo, string: str) -> dict[str, str]:
        valid_keys = {'description', 'core_version', 'python_version'}
        data = dict.fromkeys(valid_keys)
        for opt in shlex.split(string):
            if '=' not in opt:
                data[opt] = None
                continue
            key, _sep, value = opt.partition('=')
            data[key] = value
        if not any((data['core_version'], data['python_version'])):
            self.add_message(
                'ansible-deprecated-version-comment-missing-version',
                line=token.start[0],
                col_offset=token.start[1],
            )
        bad = set(data).difference(valid_keys)
        if bad:
            self.add_message(
                'ansible-deprecated-version-comment-invalid-key',
                line=token.start[0],
                col_offset=token.start[1],
                args=(','.join(bad),),
            )
        return data

    def _process_python_version(self, token: tokenize.TokenInfo, data: dict[str, str]) -> None:
        check_version = '.'.join(map(str, self.linter.config.py_version))  # minimum supported Python version provided by ansible-test

        try:
            if LooseVersion(check_version) > LooseVersion(data['python_version']):
                self.add_message(
                    'ansible-deprecated-python-version-comment',
                    line=token.start[0],
                    col_offset=token.start[1],
                    args=(
                        data['python_version'],
                        data['description'] or 'description not provided',
                    ),
                )
        except (ValueError, TypeError) as exc:
            self.add_message(
                'ansible-deprecated-version-comment-invalid-version',
                line=token.start[0],
                col_offset=token.start[1],
                args=(data['python_version'], exc),
            )

    def _process_core_version(self, token: tokenize.TokenInfo, data: dict[str, str]) -> None:
        try:
            if self.ANSIBLE_VERSION >= LooseVersion(data['core_version']):
                self.add_message(
                    'ansible-deprecated-version-comment',
                    line=token.start[0],
                    col_offset=token.start[1],
                    args=(
                        data['core_version'],
                        data['description'] or 'description not provided',
                    ),
                )
        except (ValueError, TypeError) as exc:
            self.add_message(
                'ansible-deprecated-version-comment-invalid-version',
                line=token.start[0],
                col_offset=token.start[1],
                args=(data['core_version'], exc),
            )

    def _process_comment(self, token: tokenize.TokenInfo) -> None:
        if token.string.startswith('# deprecated:'):
            data = self._deprecated_string_to_dict(token, token.string[13:].strip())
            if data['core_version']:
                self._process_core_version(token, data)
            if data['python_version']:
                self._process_python_version(token, data)


def register(linter: pylint.lint.PyLinter) -> None:
    """Required method to auto-register this checker."""
    linter.register_checker(AnsibleDeprecatedCommentChecker(linter))



"""Legacy wire format for module to controller communication."""

from __future__ import annotations as _annotations

import datetime as _datetime

from .. import _profiles
from ansible.module_utils.common.text.converters import to_text as _to_text


class _Profile(_profiles._JSONSerializationProfile["Encoder", "Decoder"]):
    @classmethod
    def bytes_to_text(cls, value: bytes) -> str:
        return _to_text(value, errors='surrogateescape')

    @classmethod
    def post_init(cls) -> None:
        cls.allowed_ansible_serializable_types = _profiles._common_module_types | _profiles._common_module_response_types

        cls.serialize_map = {
            bytes: cls.bytes_to_text,  # legacy behavior from jsonify and container_to_text
            set: cls.serialize_as_list,  # legacy _json_encode_fallback behavior
            tuple: cls.serialize_as_list,  # JSONEncoder built-in behavior
            _datetime.date: cls.serialize_as_isoformat,  # legacy parameters.py does this before serialization
            _datetime.time: cls.serialize_as_isoformat,  # always failed pre-2.18, so okay to include for consistency
            _datetime.datetime: cls.serialize_as_isoformat,  # legacy _json_encode_fallback behavior *and* legacy parameters.py does this before serialization
        }

        cls.handle_key = cls._handle_key_str_fallback  # type: ignore[method-assign]  # legacy stdlib-compatible key behavior


class Encoder(_profiles.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_profiles.AnsibleProfileJSONDecoder):
    _profile = _Profile



from __future__ import annotations

from ansible.plugins.action import ActionBase
from ansible.module_utils.common.warnings import deprecate

from ..module_utils.shared_deprecation import get_deprecation_kwargs, get_deprecated_value


class ActionModule(ActionBase):
    def run(self, tmp=None, task_vars=None):
        result = super().run(tmp, task_vars)

        for deprecate_kw in get_deprecation_kwargs():
            deprecate(**deprecate_kw)  # pylint: disable=ansible-deprecated-no-version

        result.update(deprecated_result=get_deprecated_value())

        return result



"""Common classification code used by multiple languages."""

from __future__ import annotations

import os

from ..data import (
    data_context,
)


def resolve_csharp_ps_util(import_name: str, path: str) -> str:
    """Return the fully qualified name of the given import if possible, otherwise return the original import name."""
    if data_context().content.is_ansible or not import_name.startswith('.'):
        # We don't support relative paths for builtin utils, there's no point.
        return import_name

    packages = import_name.split('.')
    module_packages = path.split(os.path.sep)

    for package in packages:
        if not module_packages or package:
            break
        del module_packages[-1]

    return 'ansible_collections.%s%s' % (data_context().content.prefix,
                                         '.'.join(module_packages + [p for p in packages if p]))



# (c) 2017, Toshio Kuratomi <tkuratomi@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os.path

import pytest

import ansible.errors

from ansible.executor import module_common as amc
from ansible.executor.interpreter_discovery import InterpreterDiscoveryRequiredError


class TestStripComments:
    def test_no_changes(self):
        no_comments = u"""def some_code():
    return False"""
        assert amc._strip_comments(no_comments) == no_comments

    def test_all_comments(self):
        all_comments = u"""# This is a test
            # Being as it is
            # To be
            """
        assert amc._strip_comments(all_comments) == u"\n\n\n"

    def test_all_whitespace(self):
        all_whitespace = (
            '\n'
            '              \n'
            '\n'
            '                \n'
            '\t\t\r\n'
            '\n'
            '            '
        )

        assert amc._strip_comments(all_whitespace) == u"\n\n\n\n\n\n"

    def test_somewhat_normal(self):
        mixed = u"""#!/usr/bin/python

# here we go
def test(arg):
    # this is a thing
    thing = '# test'
    return thing
# End
"""
        mixed_results = u"""


def test(arg):

    thing = '# test'
    return thing
"""
        assert amc._strip_comments(mixed) == mixed_results


class TestSlurp:
    def test_slurp_nonexistent(self, mocker):
        mocker.patch('os.path.exists', side_effect=lambda x: False)
        with pytest.raises(ansible.errors.AnsibleError):
            amc._slurp('no_file')

    def test_slurp_file(self, mocker):
        mocker.patch('os.path.exists', side_effect=lambda x: True)
        m = mocker.mock_open(read_data='This is a test')
        mocker.patch('builtins.open', m)
        assert amc._slurp('some_file') == 'This is a test'

    def test_slurp_file_with_newlines(self, mocker):
        mocker.patch('os.path.exists', side_effect=lambda x: True)
        m = mocker.mock_open(read_data='#!/usr/bin/python\ndef test(args):\nprint("hi")\n')
        mocker.patch('builtins.open', m)
        assert amc._slurp('some_file') == '#!/usr/bin/python\ndef test(args):\nprint("hi")\n'


class TestGetShebang:
    """Note: We may want to change the API of this function in the future.  It isn't a great API"""
    def test_no_interpreter_set(self, templar):
        # normally this would return /usr/bin/python, but so long as we're defaulting to auto python discovery, we'll get
        # an InterpreterDiscoveryRequiredError here instead
        with pytest.raises(InterpreterDiscoveryRequiredError):
            amc._get_shebang(u'/usr/bin/python', {}, templar)

    def test_python_interpreter(self, templar):
        assert amc._get_shebang(u'/usr/bin/python3.8', {}, templar) == ('#!/usr/bin/python3.8', u'/usr/bin/python3.8')

    def test_non_python_interpreter(self, templar):
        assert amc._get_shebang(u'/usr/bin/ruby', {}, templar) == ('#!/usr/bin/ruby', u'/usr/bin/ruby')

    def test_interpreter_set_in_task_vars(self, templar):
        assert amc._get_shebang(u'/usr/bin/python', {u'ansible_python_interpreter': u'/usr/bin/pypy'}, templar) == \
            (u'#!/usr/bin/pypy', u'/usr/bin/pypy')

    def test_non_python_interpreter_in_task_vars(self, templar):
        assert amc._get_shebang(u'/usr/bin/ruby', {u'ansible_ruby_interpreter': u'/usr/local/bin/ruby'}, templar) == \
            (u'#!/usr/local/bin/ruby', u'/usr/local/bin/ruby')

    def test_with_args(self, templar):
        assert amc._get_shebang(u'/usr/bin/python', {u'ansible_python_interpreter': u'/usr/bin/python3'}, templar, args=('-tt', '-OO')) == \
            (u'#!/usr/bin/python3 -tt -OO', u'/usr/bin/python3')

    def test_python_via_env(self, templar):
        assert amc._get_shebang(u'/usr/bin/python', {u'ansible_python_interpreter': u'/usr/bin/env python'}, templar) == \
            (u'#!/usr/bin/env python', u'/usr/bin/env python')


class TestDetectionRegexes:
    ANSIBLE_MODULE_UTIL_STRINGS = (
        # Absolute collection imports
        b'import ansible_collections.my_ns.my_col.plugins.module_utils.my_util',
        b'from ansible_collections.my_ns.my_col.plugins.module_utils import my_util',
        b'from ansible_collections.my_ns.my_col.plugins.module_utils.my_util import my_func',
        # Absolute core imports
        b'import ansible.module_utils.basic',
        b'from ansible.module_utils import basic',
        b'from ansible.module_utils.basic import AnsibleModule',
        # Relative imports
        b'from ..module_utils import basic',
        b'from .. module_utils import basic',
        b'from ....module_utils import basic',
        b'from ..module_utils.basic import AnsibleModule',
    )
    NOT_ANSIBLE_MODULE_UTIL_STRINGS = (
        b'from ansible import release',
        b'from ..release import __version__',
        b'from .. import release',
        b'from ansible.modules.system import ping',
        b'from ansible_collecitons.my_ns.my_col.plugins.modules import function',
    )

    OFFSET = os.path.dirname(os.path.dirname(amc.__file__))
    CORE_PATHS = (
        ('%s/modules/from_role.py' % OFFSET, 'ansible/modules/from_role'),
        ('%s/modules/system/ping.py' % OFFSET, 'ansible/modules/system/ping'),
        ('%s/modules/cloud/amazon/s3.py' % OFFSET, 'ansible/modules/cloud/amazon/s3'),
    )

    COLLECTION_PATHS = (
        ('/root/ansible_collections/ns/col/plugins/modules/ping.py',
         'ansible_collections/ns/col/plugins/modules/ping'),
        ('/root/ansible_collections/ns/col/plugins/modules/subdir/ping.py',
         'ansible_collections/ns/col/plugins/modules/subdir/ping'),
    )

    @pytest.mark.parametrize('testcase', ANSIBLE_MODULE_UTIL_STRINGS)
    def test_detect_new_style_python_module_re(self, testcase):
        assert amc.NEW_STYLE_PYTHON_MODULE_RE.search(testcase)

    @pytest.mark.parametrize('testcase', NOT_ANSIBLE_MODULE_UTIL_STRINGS)
    def test_no_detect_new_style_python_module_re(self, testcase):
        assert not amc.NEW_STYLE_PYTHON_MODULE_RE.search(testcase)

    @pytest.mark.parametrize('testcase, result', CORE_PATHS)
    def test_detect_core_library_path_re(self, testcase, result):
        assert amc.CORE_LIBRARY_PATH_RE.search(testcase).group('path') == result

    @pytest.mark.parametrize('testcase', (p[0] for p in COLLECTION_PATHS))
    def test_no_detect_core_library_path_re(self, testcase):
        assert not amc.CORE_LIBRARY_PATH_RE.search(testcase)

    @pytest.mark.parametrize('testcase, result', COLLECTION_PATHS)
    def test_detect_collection_path_re(self, testcase, result):
        assert amc.COLLECTION_PATH_RE.search(testcase).group('path') == result

    @pytest.mark.parametrize('testcase', (p[0] for p in CORE_PATHS))
    def test_no_detect_collection_path_re(self, testcase):
        assert not amc.COLLECTION_PATH_RE.search(testcase)



# Copyright (c) 2024 ShIRann Chen <shirannx@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations


class ModuleDocFragment(object):

    DOCUMENTATION = r"""
options:
    checksum_algorithm:
        description:
        - Algorithm to determine checksum of file.
        - Will throw an error if the host is unable to use specified algorithm.
        - The remote host has to support the hashing method specified, V(md5)
            can be unavailable if the host is FIPS-140 compliant.
        - Availability might be restricted by the target system, for example FIPS systems won't allow md5 use
        type: str
        choices: [ md5, sha1, sha224, sha256, sha384, sha512 ]
        default: sha1
        aliases: [ checksum, checksum_algo ]
        version_added: "2.0"
    get_checksum:
        description:
        - Whether to return a checksum of the file.
        type: bool
        default: yes
"""



# -*- coding: utf-8 -*-
# (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# pylint: disable=ansible-deprecated-no-version  # arg-splatting prevents pylint from understanding how deprecate is being called

from __future__ import annotations

import typing as t

import pytest

from ansible.module_utils._internal import _traceback, _messages, _deprecator
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common import warnings
from ansible.module_utils.common.warnings import deprecate
from ansible.module_utils.testing import patch_module_args
from units.mock.module import ModuleEnvMocker

pytestmark = pytest.mark.usefixtures("as_target", "module_env_mocker")


def test_deprecate() -> None:
    deprecate('Deprecation message')
    assert warnings.get_deprecation_messages() == (dict(msg='Deprecation message', collection_name=None, version=None),)
    assert warnings.get_deprecations() == [_messages.DeprecationSummary(
        event=_messages.Event(msg='Deprecation message'),
        deprecator=_deprecator.INDETERMINATE_DEPRECATOR,
    )]


def test_deprecate_via_module() -> None:
    with patch_module_args():
        am = AnsibleModule(argument_spec={})

    am.deprecate('Deprecation message')

    assert warnings.get_deprecation_messages() == (dict(msg='Deprecation message', collection_name=None, version=None),)
    assert warnings.get_deprecations() == [_messages.DeprecationSummary(
        event=_messages.Event(msg='Deprecation message'),
        deprecator=_deprecator.INDETERMINATE_DEPRECATOR,
    )]


def test_dedupe_with_traceback(module_env_mocker: ModuleEnvMocker) -> None:
    module_env_mocker.set_traceback_config([_traceback.TracebackEvent.DEPRECATED])
    deprecate_args: dict[str, t.Any] = dict(msg="same", version="1.2.3", collection_name="blar.blar")

    # DeprecationSummary dataclass object hash is the dedupe key; presence of differing tracebacks or SourceContexts affects de-dupe

    for _i in range(0, 10):
        # same location, same traceback- should be collapsed to one message
        deprecate(**deprecate_args)

    assert len(warnings.get_deprecation_messages()) == 1
    assert len(warnings.get_deprecations()) == 1

    for _i in range(0, 10):
        deprecate(**deprecate_args)  # with tracebacks on, we should have a different source location than the first loop, but still de-dupe

    assert len(warnings.get_deprecation_messages()) == 2
    assert len(warnings.get_deprecations()) == 2


@pytest.mark.parametrize(
    'test_case',
    (
        1,
        True,
        [1],
        {'k1': 'v1'},
        (1, 2),
        6.62607004,
        b'bytestr',
        None,
    )
)
def test_deprecate_failure(test_case):
    with pytest.raises(TypeError, match=f"must be <class 'str'> instead of {type(test_case)}"):
        deprecate(test_case)  # pylint: disable=ansible-deprecated-no-version



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.galaxy.collection import _extract_tar_dir


@pytest.fixture
def fake_tar_obj(mocker):
    m_tarfile = mocker.Mock()
    m_tarfile.type = mocker.Mock(return_value=b'99')
    m_tarfile.SYMTYPE = mocker.Mock(return_value=b'22')

    return m_tarfile


def test_extract_tar_dir_exists(mocker, fake_tar_obj):
    mocker.patch('os.makedirs', return_value=None)
    m_makedir = mocker.patch('os.mkdir', return_value=None)
    mocker.patch('os.path.isdir', return_value=True)

    _extract_tar_dir(fake_tar_obj, '/some/dir', b'/some/dest')

    assert not m_makedir.called


def test_extract_tar_dir_does_not_exist(mocker, fake_tar_obj):
    mocker.patch('os.makedirs', return_value=None)
    m_makedir = mocker.patch('os.mkdir', return_value=None)
    mocker.patch('os.path.isdir', return_value=False)

    _extract_tar_dir(fake_tar_obj, '/some/dir', b'/some/dest')

    assert m_makedir.called
    assert m_makedir.call_args[0] == (b'/some/dir', 0o0755)



# Copyright 2015 Marius Gedminas <marius@gedmin.as>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# for __setstate__/__getstate__ tests

from __future__ import annotations

import unittest


from ansible.inventory.group import Group
from ansible.inventory.host import Host


class TestHost(unittest.TestCase):
    ansible_port = 22

    def setUp(self):
        self.hostA = Host('a')
        self.hostB = Host('b')

    def test_equality(self):
        self.assertEqual(self.hostA, self.hostA)
        self.assertNotEqual(self.hostA, self.hostB)
        self.assertNotEqual(self.hostA, Host('a'))

    def test_hashability(self):
        # equality implies the hash values are the same
        self.assertEqual(hash(self.hostA), hash(Host('a')))

    def test_get_vars(self):
        host_vars = self.hostA.get_vars()
        self.assertIsInstance(host_vars, dict)

    def test_repr(self):
        host_repr = repr(self.hostA)
        self.assertIsInstance(host_repr, str)

    def test_add_group(self):
        group = Group('some_group')
        group_len = len(self.hostA.groups)
        self.hostA.add_group(group)
        self.assertEqual(len(self.hostA.groups), group_len + 1)

    def test_get_groups(self):
        group = Group('some_group')
        self.hostA.add_group(group)
        groups = self.hostA.get_groups()
        self.assertEqual(len(groups), 1)
        for _group in groups:
            self.assertIsInstance(_group, Group)

    def test_equals_none(self):
        other = None
        assert not (self.hostA == other)
        assert not (other == self.hostA)
        assert self.hostA != other
        assert other != self.hostA
        self.assertNotEqual(self.hostA, other)


class TestHostWithPort(TestHost):
    ansible_port = 8822

    def setUp(self):
        self.hostA = Host(name='a', port=self.ansible_port)
        self.hostB = Host(name='b', port=self.ansible_port)

    def test_get_vars_ansible_port(self):
        host_vars = self.hostA.get_vars()
        self.assertEqual(host_vars['ansible_port'], self.ansible_port)



from __future__ import annotations

import collections.abc as c
import typing as t

import pytest

from ansible._internal._errors import _error_utils
from ansible.module_utils._internal import _messages
from units.mock.error_helper import raise_exceptions


class _TestContributesError(Exception, _error_utils.ContributesToTaskResult):
    @property
    def result_contribution(self) -> c.Mapping[str, object]:
        return dict(some_flag=True)


class _TestContributesUnreachable(Exception, _error_utils.ContributesToTaskResult):
    @property
    def omit_failed_key(self) -> bool:
        return True

    @property
    def result_contribution(self) -> c.Mapping[str, object]:
        return dict(unreachable=True)


class _TestContributesMsg(Exception, _error_utils.ContributesToTaskResult):
    @property
    def result_contribution(self) -> c.Mapping[str, object]:
        return dict(msg="contributed msg")


@pytest.mark.parametrize("exceptions,expected", (
    (
        (Exception("e0"), _TestContributesError("e1"), ValueError("e2")),
        dict(failed=True, some_flag=True, msg="e0: e1: e2"),
    ),
    (
        (Exception("e0"), ValueError("e1"), _TestContributesError("e2")),
        dict(failed=True, some_flag=True, msg="e0: e1: e2"),
    ),
    (
        (Exception("e0"), _TestContributesUnreachable("e1")),
        dict(unreachable=True, msg="e0: e1"),
    ),
    (
        (Exception("e0"), _TestContributesMsg()),
        dict(failed=True, msg="contributed msg"),
    ),
))
def test_exception_result_contribution(exceptions: t.Sequence[BaseException], expected: dict[str, t.Any]) -> None:
    """Validate result dict augmentation by exceptions conforming to the ContributeToTaskResult protocol."""

    with pytest.raises(Exception) as error:
        raise_exceptions(exceptions)

    result = _error_utils.result_dict_from_exception(error.value, accept_result_contribution=True)

    summary = result.pop('exception')

    assert isinstance(summary, _messages.ErrorSummary)
    assert result == expected



"""Common code for source providers."""

from __future__ import annotations

import abc

from .. import (
    PathProvider,
)


class SourceProvider(PathProvider):
    """Base class for source providers."""

    @abc.abstractmethod
    def get_paths(self, path: str) -> list[str]:
        """Return the list of available content paths under the given path."""



from __future__ import annotations as _annotations

import datetime as _datetime
import os as _os
import pwd as _pwd
import time as _time

from ansible import constants as _constants
from ansible.module_utils._internal import _datatag


def generate_ansible_template_vars(
    path: str,
    fullpath: str | None = None,
    dest_path: str | None = None,
    include_ansible_managed: bool = True,
) -> dict[str, object]:
    """
    Generate and return a dictionary with variable metadata about the template specified by `fullpath`.
    If `fullpath` is `None`, `path` will be used instead.
    """
    # deprecated description="update the ansible.windows collection to inline this logic instead of calling this internal function" core_version="2.23"
    if fullpath is None:
        fullpath = _os.path.abspath(path)

    template_path = fullpath
    template_stat = _os.stat(template_path)

    template_uid: int | str

    try:
        template_uid = _pwd.getpwuid(template_stat.st_uid).pw_name
    except KeyError:
        template_uid = template_stat.st_uid

    temp_vars = dict(
        template_host=_os.uname()[1],
        template_path=path,
        template_mtime=_datetime.datetime.fromtimestamp(template_stat.st_mtime),
        template_uid=template_uid,
        template_run_date=_datetime.datetime.now(),
        template_destpath=dest_path,
        template_fullpath=fullpath,
    )

    if include_ansible_managed:  # only inject the config default value if the variable wasn't set
        temp_vars['ansible_managed'] = _generate_ansible_managed(template_stat)

    return temp_vars


def _generate_ansible_managed(template_stat: _os.stat_result) -> str:
    """Generate and return the `ansible_managed` variable."""
    # deprecated description="remove the `_generate_ansible_managed` function and use a constant instead" core_version="2.23"

    from ansible.template import trust_as_template

    managed_default = _constants.config.get_config_value('DEFAULT_MANAGED_STR')

    managed_str = managed_default.format(
        # IMPORTANT: These values must be constant strings to avoid template injection.
        #            Use Jinja template expressions where variables are needed.
        host="{{ template_host }}",
        uid="{{ template_uid }}",
        file="{{ template_path }}",
    )

    ansible_managed = _time.strftime(managed_str, _time.localtime(template_stat.st_mtime))
    ansible_managed = _datatag.AnsibleTagHelper.tag_copy(managed_default, ansible_managed)
    ansible_managed = trust_as_template(ansible_managed)

    return ansible_managed



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.utils.display import Display


@pytest.fixture()
def display_resource(request):
    Display._Singleton__instance = None

    def teardown():
        Display._Singleton__instance = None

    request.addfinalizer(teardown)



# -*- coding: utf-8 -*-
# Copyright: (c) 2017 Ansible Project
# License: GNU General Public License v3 or later (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt )

from __future__ import annotations

import pytest

from ansible.module_utils.parsing.convert_bool import boolean


junk_values = ("flibbity", 42, 42.0, object(), None, 2, -1, 0.1)


@pytest.mark.parametrize(("value", "expected"), [
    (True, True),
    (False, False),
    (1, True),
    (0, False),
    (0.0, False),
    ("true", True),
    ("TRUE", True),
    ("t", True),
    ("yes", True),
    ("y", True),
    ("on", True),
])
def test_boolean(value: object, expected: bool) -> None:
    assert boolean(value) is expected


@pytest.mark.parametrize("value", junk_values)
def test_junk_values_non_strict(value: object) -> None:
    assert boolean(value, strict=False) is False


@pytest.mark.parametrize("value", junk_values)
def test_junk_values_strict(value: object) -> None:
    with pytest.raises(TypeError, match=f"^The value '{value}' is not"):
        boolean(value, strict=True)



# -*- coding: utf-8 -*-
# (c) 2018 Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.module_utils.urls import generic_urlparse
from urllib.parse import urlparse, urlunparse


def test_generic_urlparse():
    url = 'https://ansible.com/blog'
    parts = urlparse(url)
    generic_parts = generic_urlparse(parts)
    assert generic_parts.as_list() == list(parts)

    assert urlunparse(generic_parts.as_list()) == url


def test_generic_urlparse_netloc():
    url = 'https://ansible.com:443/blog'
    parts = urlparse(url)
    generic_parts = generic_urlparse(parts)
    assert generic_parts.hostname == parts.hostname
    assert generic_parts.hostname == 'ansible.com'
    assert generic_parts.port == 443
    assert urlunparse(generic_parts.as_list()) == url



# -*- coding: utf-8 -*-
# Copyright The Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

from ansible.playbook.attribute import FieldAttribute


class Delegatable:
    delegate_to = FieldAttribute(isa='string')
    delegate_facts = FieldAttribute(isa='bool')

    def _post_validate_delegate_to(self, attr, value, templar):
        """This method exists just to make it clear that ``Task.post_validate``
        does not template this value, it is set via ``TaskExecutor._calculate_delegate_to``
        """
        return value



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import typing as t

from ansible.errors import AnsibleError
from ansible.module_utils.common.sentinel import Sentinel
from ansible.module_utils._internal._datatag import AnsibleTagHelper
from ansible.playbook.attribute import FieldAttribute
from ansible._internal._templating._engine import TemplateEngine


def _flatten_tags(tags: list[str | int]) -> list[str | int]:
    rv = set()
    for tag in tags:
        if isinstance(tag, list):
            rv.update(tag)
        else:
            rv.add(tag)
    return list(rv)


class Taggable:

    untagged = frozenset(['untagged'])
    tags = FieldAttribute(isa='list', default=list, listof=(str, int), extend=True)

    def _load_tags(self, attr, ds):
        if isinstance(ds, list):
            return ds

        if isinstance(ds, str):
            return [AnsibleTagHelper.tag_copy(ds, item.strip()) for item in ds.split(',')]

        raise AnsibleError('tags must be specified as a list', obj=ds)

    def _get_all_taggable_objects(self) -> t.Iterable[Taggable]:
        obj = self
        while obj is not None:
            yield obj

            if (role := getattr(obj, "_role", Sentinel)) is not Sentinel:
                yield role  # type: ignore[misc]

            obj = obj._parent

        yield self.get_play()

    def evaluate_tags(self, only_tags, skip_tags, all_vars):
        """Check if the current item should be executed depending on the specified tags.

        NOTE this method is assumed to be called only on Task objects.
        """
        if self.tags:
            templar = TemplateEngine(loader=self._loader, variables=all_vars)
            for obj in self._get_all_taggable_objects():
                if (_tags := getattr(obj, "_tags", Sentinel)) is not Sentinel:
                    obj._tags = _flatten_tags(templar.template(_tags))
            tags = set(self.tags)
        else:
            # this makes isdisjoint work for untagged
            tags = self.untagged

        should_run = True  # default, tasks to run

        if only_tags:
            if 'always' in tags:
                should_run = True
            elif ('all' in only_tags and 'never' not in tags):
                should_run = True
            elif not tags.isdisjoint(only_tags):
                should_run = True
            elif 'tagged' in only_tags and tags != self.untagged and 'never' not in tags:
                should_run = True
            else:
                should_run = False

        if should_run and skip_tags:

            # Check for tags that we need to skip
            if 'all' in skip_tags:
                if 'always' not in tags or 'always' in skip_tags:
                    should_run = False
            elif not tags.isdisjoint(skip_tags):
                should_run = False
            elif 'tagged' in skip_tags and tags != self.untagged:
                should_run = False

        return should_run



"""PyYAML compatibility."""

from __future__ import annotations

import typing as t

from functools import (
    partial,
)

try:
    import yaml as _yaml

    YAML_IMPORT_ERROR = None
except ImportError as ex:
    yaml_load = None  # pylint: disable=invalid-name
    YAML_IMPORT_ERROR = ex
else:
    try:
        _SafeLoader: t.Union[t.Type[_yaml.CSafeLoader], t.Type[_yaml.SafeLoader]] = _yaml.CSafeLoader
    except AttributeError:
        _SafeLoader = _yaml.SafeLoader

    yaml_load = partial(_yaml.load, Loader=_SafeLoader)



# Copyright (c) 2022 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


DOCUMENTATION = """
    name: broken_docs
    -
"""

EXAMPLE = """
"""

RETURN = """
"""



from __future__ import annotations

from ansible.plugins.action import ActionBase


class ActionModule(ActionBase):
    def run(self, tmp=None, task_vars=None):
        return dict(changed=False, failed=False, msg="action result should be masked", _ansible_no_log="yeppers")  # ensure that a truthy non-bool works here



from __future__ import annotations

import typing as t

from ansible._internal._datatag._tags import TrustedAsTemplate


def apply_trust(value: object) -> object:
    """
    Filter that returns a tagged copy of the input string with TrustedAsTemplate.
    Containers and other non-string values are returned unmodified.
    """
    return TrustedAsTemplate().tag(value) if isinstance(value, str) else value


class FilterModule:
    @staticmethod
    def filters() -> dict[str, t.Callable]:
        return dict(apply_trust=apply_trust)



# -*- coding: utf-8 -*-

# Copyright: (c) 2016, Ansible RedHat, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


DOCUMENTATION = r"""
---
module: set_stats
short_description: Define and display stats for the current ansible run
description:
     - This module allows setting/accumulating stats on the current ansible run, either per host or for all hosts in the run.
     - This module is also supported for Windows targets.
author: Brian Coca (@bcoca)
options:
  data:
    description:
      - A dictionary of which each key represents a stat (or variable) you want to keep track of.
    type: dict
    required: true
  per_host:
    description:
        - Whether the stats are per host or for all hosts in the run.
    type: bool
    default: no
  aggregate:
    description:
        - Whether the provided value is aggregated to the existing stat V(true) or will replace it V(false).
    type: bool
    default: yes
extends_documentation_fragment:
    - action_common_attributes
    - action_common_attributes.conn
    - action_common_attributes.flow
    - action_core
attributes:
    action:
        details: While the action plugin does do some of the work it relies on the core engine to actually create the variables, that part cannot be overridden
        support: partial
    bypass_host_loop:
        support: none
    bypass_task_loop:
        support: none
    core:
        details: While parts of this action are implemented in core, other parts are still available as normal plugins and can be partially overridden
        support: partial
    check_mode:
        support: full
    delegation:
        support: none
    diff_mode:
        support: none
notes:
    - In order for custom stats to be displayed, you must set C(show_custom_stats) in section C([defaults]) in C(ansible.cfg)
      or by defining environment variable C(ANSIBLE_SHOW_CUSTOM_STATS) to V(true). See the P(ansible.builtin.default#callback) callback plugin for details.
version_added: "2.3"
"""

EXAMPLES = r"""
- name: Aggregating packages_installed stat per host
  ansible.builtin.set_stats:
    data:
      packages_installed: 31
    per_host: yes

- name: Aggregating random stats for all hosts using complex arguments
  ansible.builtin.set_stats:
    data:
      one_stat: 11
      other_stat: "{{ local_var * 2 }}"
      another_stat: "{{ some_registered_var.results | map(attribute='ansible_facts.some_fact') | list }}"
    per_host: no

- name: Setting stats (not aggregating)
  ansible.builtin.set_stats:
    data:
      the_answer: 42
    aggregate: no
"""



from __future__ import annotations

import contextlib
import re
import typing as t

from ansible._internal import _display_utils
from ansible.module_utils._internal import _messages


@contextlib.contextmanager
def emits_warnings(
    warning_pattern: list[str] | str | None = None,
    deprecation_pattern: list[str] | str | None = None,
    ignore_boilerplate: bool = True,
    allow_unmatched_message: bool = False,
) -> t.Iterator[None]:
    """Assert that the code within the context manager body emits a warning or deprecation warning whose formatted output matches the supplied regex."""
    with _display_utils.DeferredWarningContext(variables=dict(ansible_deprecation_warnings=True)) as ctx:
        yield

    deprecations = ctx.get_deprecation_warnings()
    warnings = ctx.get_warnings()

    if ignore_boilerplate:
        warnings = [warning for warning in warnings if not warning.event.msg.startswith('Deprecation warnings can be disabled by setting')]

    _check_messages('Warning', warning_pattern, warnings, allow_unmatched_message)
    _check_messages('Deprecation', deprecation_pattern, deprecations, allow_unmatched_message)


def _check_messages(
    label: str,
    patterns: list[str] | str | None,
    entries: list[_messages.WarningSummary] | list[_messages.DeprecationSummary],
    allow_unmatched_message: bool,
) -> None:
    if patterns is None:
        return

    if isinstance(patterns, str):
        patterns = [patterns]

    unmatched = set(str(entry) for entry in entries)

    for pattern in patterns:
        matched = False

        for entry in entries:
            str_entry = str(entry)

            if re.search(pattern, str_entry):
                unmatched.discard(str_entry)
                matched = True

        assert matched, f"{label} pattern {pattern!r} did not match."

    if not allow_unmatched_message:
        assert not unmatched, f"{label} unmatched: {unmatched}"



from __future__ import annotations

import functools
import json
import json.encoder
import json.decoder
import typing as t

from .._wrapt import ObjectProxy
from .._json._profiles import _cache_persistence


class PluginInterposer(ObjectProxy):
    """Proxies a Cache plugin instance to implement transparent encapsulation of serialized Ansible internal data types."""

    _PAYLOAD_KEY = '__payload__'
    """The key used to store the serialized payload."""

    def get(self, key: str) -> dict[str, object]:
        return self._decode(self.__wrapped__.get(self._get_key(key)))

    def set(self, key: str, value: dict[str, object]) -> None:
        self.__wrapped__.set(self._get_key(key), self._encode(value))

    def keys(self) -> t.Sequence[str]:
        return [k for k in (self._restore_key(k) for k in self.__wrapped__.keys()) if k is not None]

    def contains(self, key: t.Any) -> bool:
        return self.__wrapped__.contains(self._get_key(key))

    def delete(self, key: str) -> None:
        self.__wrapped__.delete(self._get_key(key))

    @classmethod
    def _restore_key(cls, wrapped_key: str) -> str | None:
        prefix = cls._get_wrapped_key_prefix()

        if not wrapped_key.startswith(prefix):
            return None

        return wrapped_key[len(prefix) :]

    @classmethod
    @functools.cache
    def _get_wrapped_key_prefix(cls) -> str:
        return f's{_cache_persistence._Profile.schema_id}_'

    @classmethod
    def _get_key(cls, key: str) -> str:
        """Augment the supplied key with a schema identifier to allow for side-by-side caching across incompatible schemas."""
        return f'{cls._get_wrapped_key_prefix()}{key}'

    def _encode(self, value: dict[str, object]) -> dict[str, object]:
        return {self._PAYLOAD_KEY: json.dumps(value, cls=_cache_persistence.Encoder)}

    def _decode(self, value: dict[str, t.Any]) -> dict[str, object]:
        return json.loads(value[self._PAYLOAD_KEY], cls=_cache_persistence.Decoder)



"""
This is testing JSON functionality from module_utils which is only available in a controller context.
Do not add tests here which should pass in a target context.
"""

from __future__ import annotations

import json
import pytest

from ansible.errors import AnsibleRuntimeError
from ansible.module_utils._internal._json._profiles import _tagless


@pytest.mark.parametrize("value", (
    r'"\ud8f3"',
    r'["\ud8f3"]',
    r'[["\ud8f3"]]',
    r'[[{"key": "\ud8f3"}]]',
    r'{"key": "\ud8f3"}',
    r'{"key": ["\ud8f3"]}',
    r'{"key": {"subkey": "\ud8f3"}}',
    r'{"key": {"subkey": ["\ud8f3"]}}',
    r'{"key": [{"subkey": ["\ud8f3"]}]}',
    r'{"\ud8f3": "value"}',
    r'{"key": {"\ud8f3": "subvalue"}}',
    r'{"key": [{"\ud8f3": "subvalue"}]}',
))
def test_invalid_utf8_decoding(value: str) -> None:
    """Verify that strings which cannot be encoded as valid UTF8 result in an error during deserialization."""
    with pytest.raises(AnsibleRuntimeError, match='^Refusing to deserialize an invalid UTF8 string value'):
        json.loads(value, cls=_tagless.Decoder)



from __future__ import annotations

import json
import textwrap

import pytest

from ansible.errors import AnsibleJSONParserError
from ansible.parsing.vault import VaultHelper, EncryptedString
from ansible._internal._yaml._errors import AnsibleYAMLParserError
from ansible.parsing.utils.yaml import from_yaml

from ...mock.vault_helper import VaultTestHelper


def test_from_yaml_json_only(_vault_secrets_context: VaultTestHelper) -> None:
    """Ensure that from_yaml properly yields an `EncryptedString` instance for legacy-profile JSON with encoded vaulted values."""
    ciphertext = _vault_secrets_context.make_vault_ciphertext('mom')

    data = json.dumps(dict(hi=dict(
        __ansible_vault=ciphertext,
    )))

    result = from_yaml(data=data, file_name='/nope.json', json_only=True)

    assert isinstance(result['hi'], EncryptedString)
    assert result == dict(hi='mom')
    assert VaultHelper.get_ciphertext(result['hi'], with_tags=False) == ciphertext


def test_from_yaml(_vault_secrets_context: VaultTestHelper) -> None:
    ciphertext = _vault_secrets_context.make_vault_ciphertext('mom')

    data = f'hi: !vault |\n{textwrap.indent(ciphertext, "  ")}'

    result = from_yaml(data=data, file_name='/nope.yml')

    assert result == dict(hi='mom')
    assert VaultHelper.get_ciphertext(result['hi'], with_tags=False) == ciphertext


def test_from_yaml_invalid_json_vaulted_value() -> None:
    data = json.dumps(dict(hi=dict(__ansible_vault=1)))

    with pytest.raises(AnsibleJSONParserError, match="__ansible_vault is <class 'int'> not <class 'str'>"):
        from_yaml(data=data, json_only=True)


def test_from_yaml_invalid_yaml_vaulted_value() -> None:
    data = 'hi: !vault 1'

    with pytest.raises(AnsibleYAMLParserError, match="The '!vault' tag requires a string value."):
        from_yaml(data=data)


def test_from_yaml_does_not_recognize_json_vaulted_value() -> None:
    value = dict(hi=dict(__ansible_vault=dict(value_does_not_matter=True)))
    data = json.dumps(value)

    assert from_yaml(data=data) == value



"""Test dataclass patching to verify ClassVar works properly."""

from __future__ import annotations

import dataclasses
import typing as t

from typing import ClassVar

from ansible.module_utils.compat import typing as typing_surrogate
from ansible.module_utils.compat.typing import ClassVar as ClassVarFromSurrogate
from ansible.module_utils._internal._datatag import _tag_dataclass_kwargs


def test_classvar_fields() -> None:
    """Verify that `typing.ClassVar` works with dataclasses."""
    @dataclasses.dataclass(**_tag_dataclass_kwargs)
    class ExerciseClassVar:
        real_local: ClassVar[int]
        real_from_module: t.ClassVar[int]
        surrogate_local: ClassVarFromSurrogate
        # this is the case that was actually broken; treated as an instance field when ClassVar was dot-referenced from a non-typing module
        surrogate_from_module: typing_surrogate.ClassVar[int]

        instance_field: int = 42

    fields = dataclasses.fields(ExerciseClassVar)

    assert len(fields) == 1
    assert ExerciseClassVar()

    # ensure that the classvars are all settable; some forms of the failure prevent this
    ExerciseClassVar.real_local = 42
    ExerciseClassVar.real_from_module = 42
    ExerciseClassVar.surrogate_local = 42
    ExerciseClassVar.surrogate_from_module = 42

    assert fields[-1].name == 'instance_field'



# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Toshio Kuratomi <tkuratomi@ansible.com>
# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import datetime
import os
import pytest
import zipfile

from io import BytesIO

import ansible.errors

from ansible._internal._ansiballz._builder import ExtensionManager
from ansible.executor.module_common import recursive_finder
from ansible.plugins.loader import init_plugin_loader

# These are the modules that are brought in by module_utils/basic.py  This may need to be updated
# when basic.py gains new imports
# We will remove these when we modify AnsiBallZ to store its args in a separate file instead of in
# basic.py

MODULE_UTILS_BASIC_FILES = frozenset(('ansible/__init__.py',
                                      'ansible/module_utils/__init__.py',
                                      'ansible/module_utils/basic.py',
                                      'ansible/module_utils/six/__init__.py',
                                      'ansible/module_utils/_internal/__init__.py',
                                      'ansible/module_utils/_internal/_ansiballz/__init__.py',
                                      'ansible/module_utils/_internal/_ansiballz/_loader.py',
                                      'ansible/module_utils/_internal/_dataclass_validation.py',
                                      'ansible/module_utils/_internal/_datatag/__init__.py',
                                      'ansible/module_utils/_internal/_datatag/_tags.py',
                                      'ansible/module_utils/_internal/_debugging.py',
                                      'ansible/module_utils/_internal/_deprecator.py',
                                      'ansible/module_utils/_internal/_errors.py',
                                      'ansible/module_utils/_internal/_event_utils.py',
                                      'ansible/module_utils/_internal/_json/__init__.py',
                                      'ansible/module_utils/_internal/_json/_legacy_encoder.py',
                                      'ansible/module_utils/_internal/_json/_profiles/__init__.py',
                                      'ansible/module_utils/_internal/_json/_profiles/_module_legacy_c2m.py',
                                      'ansible/module_utils/_internal/_json/_profiles/_module_legacy_m2c.py',
                                      'ansible/module_utils/_internal/_json/_profiles/_tagless.py',
                                      'ansible/module_utils/_internal/_traceback.py',
                                      'ansible/module_utils/_internal/_validation.py',
                                      'ansible/module_utils/_internal/_messages.py',
                                      'ansible/module_utils/_internal/_patches/_dataclass_annotation_patch.py',
                                      'ansible/module_utils/_internal/_patches/_socket_patch.py',
                                      'ansible/module_utils/_internal/_patches/_sys_intern_patch.py',
                                      'ansible/module_utils/_internal/_patches/__init__.py',
                                      'ansible/module_utils/_internal/_plugin_info.py',
                                      'ansible/module_utils/_internal/_stack.py',
                                      'ansible/module_utils/_internal/_text_utils.py',
                                      'ansible/module_utils/common/collections.py',
                                      'ansible/module_utils/common/parameters.py',
                                      'ansible/module_utils/common/warnings.py',
                                      'ansible/module_utils/parsing/convert_bool.py',
                                      'ansible/module_utils/common/__init__.py',
                                      'ansible/module_utils/common/file.py',
                                      'ansible/module_utils/common/json.py',
                                      'ansible/module_utils/common/locale.py',
                                      'ansible/module_utils/common/process.py',
                                      'ansible/module_utils/common/sys_info.py',
                                      'ansible/module_utils/common/text/__init__.py',
                                      'ansible/module_utils/common/text/converters.py',
                                      'ansible/module_utils/common/text/formatters.py',
                                      'ansible/module_utils/common/validation.py',
                                      'ansible/module_utils/common/_utils.py',
                                      'ansible/module_utils/common/arg_spec.py',
                                      'ansible/module_utils/compat/__init__.py',
                                      'ansible/module_utils/compat/selinux.py',
                                      'ansible/module_utils/compat/typing.py',
                                      'ansible/module_utils/datatag.py',
                                      'ansible/module_utils/distro/__init__.py',
                                      'ansible/module_utils/distro/_distro.py',
                                      'ansible/module_utils/errors.py',
                                      'ansible/module_utils/parsing/__init__.py',
                                      'ansible/module_utils/parsing/convert_bool.py',
                                      'ansible/module_utils/six/__init__.py',
                                      ))

ONLY_BASIC_FILE = frozenset(('ansible/module_utils/basic.py',))

ANSIBLE_LIB = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__)))), 'lib', 'ansible')

NOW = datetime.datetime.now(datetime.timezone.utc)


@pytest.fixture
def zip_file() -> zipfile.ZipFile:
    init_plugin_loader()

    zipoutput = BytesIO()
    zf = zipfile.ZipFile(zipoutput, mode='w', compression=zipfile.ZIP_STORED)

    return zf


def test_no_module_utils(zip_file: zipfile.ZipFile) -> None:
    name = 'ping'
    data = b'#!/usr/bin/python\nreturn \'{\"changed\": false}\''
    recursive_finder(name, os.path.join(ANSIBLE_LIB, 'modules', 'system', 'ping.py'), data, zip_file, NOW, ExtensionManager())
    assert frozenset(zip_file.namelist()) == MODULE_UTILS_BASIC_FILES


def test_module_utils_with_syntax_error(zip_file: zipfile.ZipFile) -> None:
    name = 'fake_module'
    data = b'#!/usr/bin/python\ndef something(:\n   pass\n'
    with pytest.raises(ansible.errors.AnsibleError) as exec_info:
        recursive_finder(name, os.path.join(ANSIBLE_LIB, 'modules', 'system', 'fake_module.py'), data, zip_file, NOW, ExtensionManager())
    assert "Unable to compile 'fake_module': invalid syntax" in str(exec_info.value)


def test_module_utils_with_identation_error(zip_file: zipfile.ZipFile) -> None:
    name = 'fake_module'
    data = b'#!/usr/bin/python\n    def something():\n    pass\n'
    with pytest.raises(ansible.errors.AnsibleError) as exec_info:
        recursive_finder(name, os.path.join(ANSIBLE_LIB, 'modules', 'system', 'fake_module.py'), data, zip_file, NOW, ExtensionManager())
    assert "Unable to compile 'fake_module': unexpected indent" in str(exec_info.value)


def test_from_import_six(zip_file: zipfile.ZipFile) -> None:
    name = 'ping'
    data = b'#!/usr/bin/python\nfrom ansible.module_utils import six'
    recursive_finder(name, os.path.join(ANSIBLE_LIB, 'modules', 'system', 'ping.py'), data, zip_file, NOW, ExtensionManager())
    assert frozenset(zip_file.namelist()) == frozenset(('ansible/module_utils/six/__init__.py', )).union(MODULE_UTILS_BASIC_FILES)


def test_import_six(zip_file: zipfile.ZipFile) -> None:
    name = 'ping'
    data = b'#!/usr/bin/python\nimport ansible.module_utils.six'
    recursive_finder(name, os.path.join(ANSIBLE_LIB, 'modules', 'system', 'ping.py'), data, zip_file, NOW, ExtensionManager())
    assert frozenset(zip_file.namelist()) == frozenset(('ansible/module_utils/six/__init__.py', )).union(MODULE_UTILS_BASIC_FILES)


def test_import_six_from_many_submodules(zip_file: zipfile.ZipFile) -> None:
    name = 'ping'
    data = b'#!/usr/bin/python\nfrom ansible.module_utils.six.moves.urllib.parse import urlparse'
    recursive_finder(name, os.path.join(ANSIBLE_LIB, 'modules', 'system', 'ping.py'), data, zip_file, NOW, ExtensionManager())
    assert frozenset(zip_file.namelist()) == frozenset(('ansible/module_utils/six/__init__.py',)).union(MODULE_UTILS_BASIC_FILES)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2013, bleader
# Written by bleader <bleader@ratonland.org>
# Based on pkgin module written by Shaun Zinck <shaun.zinck at gmail.com>
# that was based on pacman module written by Afterburn <https://github.com/afterburn>
#  that was based on apt module written by Matthew Williams <matthew@flowroute.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = """
---
module: pkgng
short_description: Package manager for FreeBSD >= 9.0
description:
    - Manage binary packages for FreeBSD using 'pkgng' which is available in versions after 9.0.
version_added: "1.2"
options:
    name:
        description:
            - Name or list of names of packages to install/remove.
        required: true
    state:
        description:
            - State of the package.
            - 'Note: "latest" added in 2.7'
        choices: [ 'present', 'latest', 'absent' ]
        required: false
        default: present
    cached:
        description:
            - Use local package base instead of fetching an updated one.
        type: bool
        required: false
        default: no
    annotation:
        description:
            - A comma-separated list of keyvalue-pairs of the form
              C(<+/-/:><key>[=<value>]). A C(+) denotes adding an annotation, a
              C(-) denotes removing an annotation, and C(:) denotes modifying an
              annotation.
              If setting or modifying annotations, a value must be provided.
        required: false
        version_added: "1.6"
    pkgsite:
        description:
            - For pkgng versions before 1.1.4, specify packagesite to use
              for downloading packages. If not specified, use settings from
              C(/usr/local/etc/pkg.conf).
            - For newer pkgng versions, specify a the name of a repository
              configured in C(/usr/local/etc/pkg/repos).
        required: false
    rootdir:
        description:
            - For pkgng versions 1.5 and later, pkg will install all packages
              within the specified root directory.
            - Can not be used together with I(chroot) or I(jail) options.
        required: false
    chroot:
        version_added: "2.1"
        description:
            - Pkg will chroot in the specified environment.
            - Can not be used together with I(rootdir) or I(jail) options.
        required: false
    jail:
        version_added: "2.4"
        description:
            - Pkg will execute in the given jail name or id.
            - Can not be used together with I(chroot) or I(rootdir) options.
    autoremove:
        version_added: "2.2"
        description:
            - Remove automatically installed packages which are no longer needed.
        required: false
        type: bool
        default: no
author: "bleader (@bleader)"
notes:
  - When using pkgsite, be careful that already in cache packages won't be downloaded again.
  - When used with a `loop:` each package will be processed individually,
    it is much more efficient to pass the list directly to the `name` option.
"""

EXAMPLES = """
- name: Install package foo
  pkgng:
    name: foo
    state: present

- name: Annotate package foo and bar
  pkgng:
    name: foo,bar
    annotation: '+test1=baz,-test2,:test3=foobar'

- name: Remove packages foo and bar
  pkgng:
    name: foo,bar
    state: absent

# "latest" support added in 2.7
- name: Upgrade package baz
  pkgng:
    name: baz
    state: latest
"""


import re
from ansible.module_utils.basic import AnsibleModule


def query_package(module, pkgng_path, name, dir_arg):

    rc, out, err = module.run_command("%s %s info -g -e %s" % (pkgng_path, dir_arg, name))

    if rc == 0:
        return True

    return False


def query_update(module, pkgng_path, name, dir_arg, old_pkgng, pkgsite):

    # Check to see if a package upgrade is available.
    # rc = 0, no updates available or package not installed
    # rc = 1, updates available
    if old_pkgng:
        rc, out, err = module.run_command("%s %s upgrade -g -n %s" % (pkgsite, pkgng_path, name))
    else:
        rc, out, err = module.run_command("%s %s upgrade %s -g -n %s" % (pkgng_path, dir_arg, pkgsite, name))

    if rc == 1:
        return True

    return False


def pkgng_older_than(module, pkgng_path, compare_version):

    rc, out, err = module.run_command("%s -v" % pkgng_path)
    version = [int(x) for x in re.split(r'[\._]', out)]

    i = 0
    new_pkgng = True
    while compare_version[i] == version[i]:
        i += 1
        if i == min(len(compare_version), len(version)):
            break
    else:
        if compare_version[i] > version[i]:
            new_pkgng = False
    return not new_pkgng


def remove_packages(module, pkgng_path, packages, dir_arg):

    remove_c = 0
    # Using a for loop in case of error, we can report the package that failed
    for package in packages:
        # Query the package first, to see if we even need to remove
        if not query_package(module, pkgng_path, package, dir_arg):
            continue

        if not module.check_mode:
            rc, out, err = module.run_command("%s %s delete -y %s" % (pkgng_path, dir_arg, package))

        if not module.check_mode and query_package(module, pkgng_path, package, dir_arg):
            module.fail_json(msg="failed to remove %s: %s" % (package, out))

        remove_c += 1

    if remove_c > 0:

        return (True, "removed %s package(s)" % remove_c)

    return (False, "package(s) already absent")


def install_packages(module, pkgng_path, packages, cached, pkgsite, dir_arg, state):

    install_c = 0

    # as of pkg-1.1.4, PACKAGESITE is deprecated in favor of repository definitions
    # in /usr/local/etc/pkg/repos
    old_pkgng = pkgng_older_than(module, pkgng_path, [1, 1, 4])
    if pkgsite != "":
        if old_pkgng:
            pkgsite = "PACKAGESITE=%s" % (pkgsite)
        else:
            pkgsite = "-r %s" % (pkgsite)

    # This environment variable skips mid-install prompts,
    # setting them to their default values.
    batch_var = 'env BATCH=yes'

    if not module.check_mode and not cached:
        if old_pkgng:
            rc, out, err = module.run_command("%s %s update" % (pkgsite, pkgng_path))
        else:
            rc, out, err = module.run_command("%s %s update" % (pkgng_path, dir_arg))
        if rc != 0:
            module.fail_json(msg="Could not update catalogue [%d]: %s %s" % (rc, out, err))

    for package in packages:
        already_installed = query_package(module, pkgng_path, package, dir_arg)
        if already_installed and state == "present":
            continue

        update_available = query_update(module, pkgng_path, package, dir_arg, old_pkgng, pkgsite)
        if not update_available and already_installed and state == "latest":
            continue

        if not module.check_mode:
            if already_installed:
                action = "upgrade"
            else:
                action = "install"
            if old_pkgng:
                rc, out, err = module.run_command("%s %s %s %s -g -U -y %s" % (batch_var, pkgsite, pkgng_path, action, package))
            else:
                rc, out, err = module.run_command("%s %s %s %s %s -g -U -y %s" % (batch_var, pkgng_path, dir_arg, action, pkgsite, package))

        if not module.check_mode and not query_package(module, pkgng_path, package, dir_arg):
            module.fail_json(msg="failed to %s %s: %s" % (action, package, out), stderr=err)

        install_c += 1

    if install_c > 0:
        return (True, "added %s package(s)" % (install_c))

    return (False, "package(s) already %s" % (state))


def annotation_query(module, pkgng_path, package, tag, dir_arg):
    rc, out, err = module.run_command("%s %s info -g -A %s" % (pkgng_path, dir_arg, package))
    match = re.search(r'^\s*(?P<tag>%s)\s*:\s*(?P<value>\w+)' % tag, out, flags=re.MULTILINE)
    if match:
        return match.group('value')
    return False


def annotation_add(module, pkgng_path, package, tag, value, dir_arg):
    _value = annotation_query(module, pkgng_path, package, tag, dir_arg)
    if not _value:
        # Annotation does not exist, add it.
        rc, out, err = module.run_command('%s %s annotate -y -A %s %s "%s"'
                                          % (pkgng_path, dir_arg, package, tag, value))
        if rc != 0:
            module.fail_json(msg="could not annotate %s: %s"
                             % (package, out), stderr=err)
        return True
    elif _value != value:
        # Annotation exists, but value differs
        module.fail_json(
            mgs="failed to annotate %s, because %s is already set to %s, but should be set to %s"
            % (package, tag, _value, value))
        return False
    else:
        # Annotation exists, nothing to do
        return False


def annotation_delete(module, pkgng_path, package, tag, value, dir_arg):
    _value = annotation_query(module, pkgng_path, package, tag, dir_arg)
    if _value:
        rc, out, err = module.run_command('%s %s annotate -y -D %s %s'
                                          % (pkgng_path, dir_arg, package, tag))
        if rc != 0:
            module.fail_json(msg="could not delete annotation to %s: %s"
                             % (package, out), stderr=err)
        return True
    return False


def annotation_modify(module, pkgng_path, package, tag, value, dir_arg):
    _value = annotation_query(module, pkgng_path, package, tag, dir_arg)
    if not value:
        # No such tag
        module.fail_json(msg="could not change annotation to %s: tag %s does not exist"
                         % (package, tag))
    elif _value == value:
        # No change in value
        return False
    else:
        rc, out, err = module.run_command('%s %s annotate -y -M %s %s "%s"'
                                          % (pkgng_path, dir_arg, package, tag, value))
        if rc != 0:
            module.fail_json(msg="could not change annotation annotation to %s: %s"
                             % (package, out), stderr=err)
        return True


def annotate_packages(module, pkgng_path, packages, annotation, dir_arg):
    annotate_c = 0
    annotations = map(lambda _annotation:
                      re.match(r'(?P<operation>[\+-:])(?P<tag>\w+)(=(?P<value>\w+))?',
                               _annotation).groupdict(),
                      re.split(r',', annotation))

    operation = {
        '+': annotation_add,
        '-': annotation_delete,
        ':': annotation_modify
    }

    for package in packages:
        for _annotation in annotations:
            if operation[_annotation['operation']](module, pkgng_path, package, _annotation['tag'], _annotation['value']):
                annotate_c += 1

    if annotate_c > 0:
        return (True, "added %s annotations." % annotate_c)
    return (False, "changed no annotations")


def autoremove_packages(module, pkgng_path, dir_arg):
    rc, out, err = module.run_command("%s %s autoremove -n" % (pkgng_path, dir_arg))

    autoremove_c = 0

    match = re.search('^Deinstallation has been requested for the following ([0-9]+) packages', out, re.MULTILINE)
    if match:
        autoremove_c = int(match.group(1))

    if autoremove_c == 0:
        return False, "no package(s) to autoremove"

    if not module.check_mode:
        rc, out, err = module.run_command("%s %s autoremove -y" % (pkgng_path, dir_arg))

    return True, "autoremoved %d package(s)" % (autoremove_c)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(default="present", choices=["present", "latest", "absent"], required=False),
            name=dict(aliases=["pkg"], required=True, type='list'),
            cached=dict(default=False, type='bool'),
            annotation=dict(default="", required=False),
            pkgsite=dict(default="", required=False),
            rootdir=dict(default="", required=False, type='path'),
            chroot=dict(default="", required=False, type='path'),
            jail=dict(default="", required=False, type='str'),
            autoremove=dict(default=False, type='bool')),
        supports_check_mode=True,
        mutually_exclusive=[["rootdir", "chroot", "jail"]])

    pkgng_path = module.get_bin_path('pkg', True)

    p = module.params

    pkgs = p["name"]

    changed = False
    msgs = []
    dir_arg = ""

    if p["rootdir"] != "":
        old_pkgng = pkgng_older_than(module, pkgng_path, [1, 5, 0])
        if old_pkgng:
            module.fail_json(msg="To use option 'rootdir' pkg version must be 1.5 or greater")
        else:
            dir_arg = "--rootdir %s" % (p["rootdir"])

    if p["chroot"] != "":
        dir_arg = '--chroot %s' % (p["chroot"])

    if p["jail"] != "":
        dir_arg = '--jail %s' % (p["jail"])

    if p["state"] in ("present", "latest"):
        _changed, _msg = install_packages(module, pkgng_path, pkgs, p["cached"], p["pkgsite"], dir_arg, p["state"])
        changed = changed or _changed
        msgs.append(_msg)

    elif p["state"] == "absent":
        _changed, _msg = remove_packages(module, pkgng_path, pkgs, dir_arg)
        changed = changed or _changed
        msgs.append(_msg)

    if p["autoremove"]:
        _changed, _msg = autoremove_packages(module, pkgng_path, dir_arg)
        changed = changed or _changed
        msgs.append(_msg)

    if p["annotation"]:
        _changed, _msg = annotate_packages(module, pkgng_path, pkgs, p["annotation"], dir_arg)
        changed = changed or _changed
        msgs.append(_msg)

    module.exit_json(changed=changed, msg=", ".join(msgs))


if __name__ == '__main__':
    main()
